from __future__ import annotations

import itertools
import re
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Iterable, Literal, Mapping, TypeVar, overload

from litestar.datastructures.cookie import Cookie
from litestar.datastructures.headers import ETag, MutableScopeHeaders
from litestar.enums import MediaType, OpenAPIMediaType
from litestar.exceptions import ImproperlyConfiguredException
from litestar.serialization import default_serializer, encode_json, encode_msgpack, get_serializer
from litestar.status_codes import HTTP_200_OK, HTTP_204_NO_CONTENT, HTTP_304_NOT_MODIFIED
from litestar.types.empty import Empty
from litestar.utils.deprecation import deprecated, warn_deprecation
from litestar.utils.helpers import get_enum_string_value

if TYPE_CHECKING:
    from typing import Optional

    from litestar.app import Litestar
    from litestar.background_tasks import BackgroundTask, BackgroundTasks
    from litestar.connection import Request
    from litestar.types import (
        HTTPResponseBodyEvent,
        HTTPResponseStartEvent,
        Receive,
        ResponseCookies,
        ResponseHeaders,
        Scope,
        Send,
        Serializer,
        TypeEncodersMap,
    )

__all__ = ("ASGIResponse", "Response")

T = TypeVar("T")

MEDIA_TYPE_APPLICATION_JSON_PATTERN = re.compile(r"^application/(?:.+\+)?json")


class ASGIResponse:
    """A low-level ASGI response class."""

    __slots__ = (
        "background",
        "body",
        "content_length",
        "encoding",
        "is_head_response",
        "status_code",
        "_encoded_cookies",
        "headers",
    )

    _should_set_content_length: ClassVar[bool] = True
    """A flag to indicate whether the content-length header should be set by default or not."""

    def __init__(
        self,
        *,
        background: BackgroundTask | BackgroundTasks | None = None,
        body: bytes | str = b"",
        content_length: int | None = None,
        cookies: Iterable[Cookie] | None = None,
        encoded_headers: Iterable[tuple[bytes, bytes]] | None = None,
        encoding: str = "utf-8",
        headers: dict[str, Any] | Iterable[tuple[str, str]] | None = None,
        is_head_response: bool = False,
        media_type: MediaType | str | None = None,
        status_code: int | None = None,
    ) -> None:
        """A low-level ASGI response class.

        Args:
            background: A background task or a list of background tasks to be executed after the response is sent.
            body: encoded content to send in the response body.
            content_length: The response content length.
            cookies: The response cookies.
            encoded_headers: The response headers.
            encoding: The response encoding.
            headers: The response headers.
            is_head_response: A boolean indicating if the response is a HEAD response.
            media_type: The response media type.
            status_code: The response status code.
        """
        body = body.encode() if isinstance(body, str) else body
        status_code = status_code or HTTP_200_OK
        self.headers = MutableScopeHeaders()

        if encoded_headers is not None:
            warn_deprecation("3.0", kind="parameter", deprecated_name="encoded_headers", alternative="headers")
            for header_name, header_value in encoded_headers:
                self.headers.add(header_name.decode("latin-1"), header_value.decode("latin-1"))

        if headers is not None:
            for k, v in headers.items() if isinstance(headers, dict) else headers:
                self.headers.add(k, v)  # pyright: ignore

        media_type = get_enum_string_value(media_type or MediaType.JSON)

        status_allows_body = (
            status_code not in {HTTP_204_NO_CONTENT, HTTP_304_NOT_MODIFIED} and status_code >= HTTP_200_OK
        )

        if content_length is None:
            content_length = len(body)

        if not status_allows_body or is_head_response:
            if body and body != b"null":
                raise ImproperlyConfiguredException(
                    "response content is not supported for HEAD responses and responses with a status code "
                    "that does not allow content (304, 204, < 200)"
                )
            body = b""
        else:
            self.headers.setdefault(
                "content-type", (f"{media_type}; charset={encoding}" if media_type.startswith("text/") else media_type)
            )

            if self._should_set_content_length:
                self.headers.setdefault("content-length", str(content_length))

        self.background = background
        self.body = body
        self.content_length = content_length
        self._encoded_cookies = tuple(
            cookie.to_encoded_header() for cookie in (cookies or ()) if not cookie.documentation_only
        )
        self.encoding = encoding
        self.is_head_response = is_head_response
        self.status_code = status_code

    @property
    @deprecated("3.0", kind="property", alternative="encode_headers()")
    def encoded_headers(self) -> list[tuple[bytes, bytes]]:
        return self.encode_headers()

    def encode_headers(self) -> list[tuple[bytes, bytes]]:
        return [*self.headers.headers, *self._encoded_cookies]

    async def after_response(self) -> None:
        """Execute after the response is sent.

        Returns:
            None
        """
        if self.background is not None:
            await self.background()

    async def start_response(self, send: Send) -> None:
        """Emit the start event of the response. This event includes the headers and status codes.

        Args:
            send: The ASGI send function.

        Returns:
            None
        """
        event: HTTPResponseStartEvent = {
            "type": "http.response.start",
            "status": self.status_code,
            "headers": self.encode_headers(),
        }
        await send(event)

    async def send_body(self, send: Send, receive: Receive) -> None:
        """Emit the response body.

        Args:
            send: The ASGI send function.
            receive: The ASGI receive function.

        Notes:
            - Response subclasses should customize this method if there is a need to customize sending data.

        Returns:
            None
        """
        event: HTTPResponseBodyEvent = {"type": "http.response.body", "body": self.body, "more_body": False}
        await send(event)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """ASGI callable of the ``Response``.

        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """
        await self.start_response(send=send)

        if self.is_head_response:
            event: HTTPResponseBodyEvent = {"type": "http.response.body", "body": b"", "more_body": False}
            await send(event)
        else:
            await self.send_body(send=send, receive=receive)

        await self.after_response()


class Response(Generic[T]):
    """Base Litestar HTTP response class, used as the basis for all other response classes."""

    __slots__ = (
        "background",
        "content",
        "cookies",
        "encoding",
        "headers",
        "media_type",
        "status_code",
        "response_type_encoders",
    )

    content: T
    type_encoders: Optional[TypeEncodersMap] = None  # noqa: UP007

    def __init__(
        self,
        content: T,
        *,
        background: BackgroundTask | BackgroundTasks | None = None,
        cookies: ResponseCookies | None = None,
        encoding: str = "utf-8",
        headers: ResponseHeaders | None = None,
        media_type: MediaType | OpenAPIMediaType | str | None = None,
        status_code: int | None = None,
        type_encoders: TypeEncodersMap | None = None,
    ) -> None:
        """Initialize the response.

        Args:
            content: A value for the response body that will be rendered into bytes string.
            status_code: An HTTP status code.
            media_type: A value for the response ``Content-Type`` header.
            background: A :class:`BackgroundTask <.background_tasks.BackgroundTask>` instance or
                :class:`BackgroundTasks <.background_tasks.BackgroundTasks>` to execute after the response is finished.
                Defaults to ``None``.
            headers: A string keyed dictionary of response headers. Header keys are insensitive.
            cookies: A list of :class:`Cookie <.datastructures.Cookie>` instances to be set under the response
                ``Set-Cookie`` header.
            encoding: The encoding to be used for the response headers.
            type_encoders: A mapping of types to callables that transform them into types supported for serialization.
        """
        self.content = content
        self.background = background
        self.cookies: list[Cookie] = (
            [Cookie(key=key, value=value) for key, value in cookies.items()]
            if isinstance(cookies, Mapping)
            else list(cookies or [])
        )
        self.encoding = encoding
        self.headers: dict[str, Any] = (
            dict(headers) if isinstance(headers, Mapping) else {h.name: h.value for h in headers or {}}
        )
        self.media_type = media_type
        self.status_code = status_code
        self.response_type_encoders = {**(self.type_encoders or {}), **(type_encoders or {})}

    @overload
    def set_cookie(self, /, cookie: Cookie) -> None: ...

    @overload
    def set_cookie(
        self,
        key: str,
        value: str | None = None,
        max_age: int | None = None,
        expires: int | None = None,
        path: str = "/",
        domain: str | None = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Literal["lax", "strict", "none"] = "lax",
    ) -> None: ...

    def set_cookie(  # type: ignore[misc]
        self,
        key: str | Cookie,
        value: str | None = None,
        max_age: int | None = None,
        expires: int | None = None,
        path: str = "/",
        domain: str | None = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Literal["lax", "strict", "none"] = "lax",
    ) -> None:
        """Set a cookie on the response. If passed a :class:`Cookie <.datastructures.Cookie>` instance, keyword
        arguments will be ignored.

        Args:
            key: Key for the cookie or a :class:`Cookie <.datastructures.Cookie>` instance.
            value: Value for the cookie, if none given defaults to empty string.
            max_age: Maximal age of the cookie before its invalidated.
            expires: Seconds from now until the cookie expires.
            path: Path fragment that must exist in the request url for the cookie to be valid. Defaults to ``/``.
            domain: Domain for which the cookie is valid.
            secure: Https is required for the cookie.
            httponly: Forbids javascript to access the cookie via ``document.cookie``.
            samesite: Controls whether a cookie is sent with cross-site requests. Defaults to ``lax``.

        Returns:
            None.
        """
        if not isinstance(key, Cookie):
            key = Cookie(
                domain=domain,
                expires=expires,
                httponly=httponly,
                key=key,
                max_age=max_age,
                path=path,
                samesite=samesite,
                secure=secure,
                value=value,
            )
        self.cookies.append(key)

    def set_header(self, key: str, value: Any) -> None:
        """Set a header on the response.

        Args:
            key: Header key.
            value: Header value.

        Returns:
            None.
        """
        self.headers[key] = value

    def set_etag(self, etag: str | ETag) -> None:
        """Set an etag header.

        Args:
            etag: An etag value.

        Returns:
            None
        """
        self.headers["etag"] = etag.to_header() if isinstance(etag, ETag) else etag

    def delete_cookie(
        self,
        key: str,
        path: str = "/",
        domain: str | None = None,
    ) -> None:
        """Delete a cookie.

        Args:
            key: Key of the cookie.
            path: Path of the cookie.
            domain: Domain of the cookie.

        Returns:
            None.
        """
        cookie = Cookie(key=key, path=path, domain=domain, expires=0, max_age=0)
        self.cookies = [c for c in self.cookies if c != cookie]
        self.cookies.append(cookie)

    def render(self, content: Any, media_type: str, enc_hook: Serializer = default_serializer) -> bytes:
        """Handle the rendering of content into a bytes string.

        Returns:
            An encoded bytes string
        """
        if isinstance(content, bytes):
            return content

        if content is Empty:
            raise RuntimeError("The `Empty` sentinel cannot be used as response content")

        try:
            if media_type.startswith("text/") and not content:
                return b""

            if isinstance(content, str):
                return content.encode(self.encoding)

            if media_type == MediaType.MESSAGEPACK:
                return encode_msgpack(content, enc_hook)

            if MEDIA_TYPE_APPLICATION_JSON_PATTERN.match(
                media_type,
            ):
                return encode_json(content, enc_hook)

            raise ImproperlyConfiguredException(f"unsupported media_type {media_type} for content {content!r}")
        except (AttributeError, ValueError, TypeError) as e:
            raise ImproperlyConfiguredException("Unable to serialize response content") from e

    def to_asgi_response(
        self,
        app: Litestar | None,
        request: Request,
        *,
        background: BackgroundTask | BackgroundTasks | None = None,
        cookies: Iterable[Cookie] | None = None,
        encoded_headers: Iterable[tuple[bytes, bytes]] | None = None,
        headers: dict[str, str] | None = None,
        is_head_response: bool = False,
        media_type: MediaType | str | None = None,
        status_code: int | None = None,
        type_encoders: TypeEncodersMap | None = None,
    ) -> ASGIResponse:
        """Create an ASGIResponse from a Response instance.

        Args:
            app: The :class:`Litestar <.app.Litestar>` application instance.
            background: Background task(s) to be executed after the response is sent.
            cookies: A list of cookies to be set on the response.
            encoded_headers: A list of already encoded headers.
            headers: Additional headers to be merged with the response headers. Response headers take precedence.
            is_head_response: Whether the response is a HEAD response.
            media_type: Media type for the response. If ``media_type`` is already set on the response, this is ignored.
            request: The :class:`Request <.connection.Request>` instance.
            status_code: Status code for the response. If ``status_code`` is already set on the response, this is
            type_encoders: A dictionary of type encoders to use for encoding the response content.

        Returns:
            An ASGIResponse instance.
        """

        if app is not None:
            warn_deprecation(
                version="2.1",
                deprecated_name="app",
                kind="parameter",
                removal_in="3.0.0",
                alternative="request.app",
            )

        headers = {**headers, **self.headers} if headers is not None else self.headers
        cookies = self.cookies if cookies is None else itertools.chain(self.cookies, cookies)

        if type_encoders:
            type_encoders = {**type_encoders, **(self.response_type_encoders or {})}
        else:
            type_encoders = self.response_type_encoders

        media_type = get_enum_string_value(self.media_type or media_type or MediaType.JSON)

        return ASGIResponse(
            background=self.background or background,
            body=self.render(self.content, media_type, get_serializer(type_encoders)),
            cookies=cookies,
            encoded_headers=encoded_headers,
            encoding=self.encoding,
            headers=headers,
            is_head_response=is_head_response,
            media_type=media_type,
            status_code=self.status_code or status_code,
        )



from __future__ import annotations

from typing import Dict
from unittest.mock import MagicMock

import pytest

from litestar import Controller, Litestar, Router, post
from litestar.config.app import ExperimentalFeatures
from litestar.dto import AbstractDTO, DTOConfig
from litestar.dto._backend import DTOBackend
from litestar.dto._codegen_backend import DTOCodegenBackend
from litestar.testing import create_test_client

from . import Model


@pytest.fixture()
def experimental_features(use_experimental_dto_backend: bool) -> list[ExperimentalFeatures] | None:
    if use_experimental_dto_backend:
        return [ExperimentalFeatures.DTO_CODEGEN]
    return None


def test_dto_defined_on_handler(
    ModelDataDTO: type[AbstractDTO], experimental_features: list[ExperimentalFeatures]
) -> None:
    @post(dto=ModelDataDTO, signature_types=[Model])
    def handler(data: Model) -> Model:
        assert data == Model(a=1, b="2")
        return data

    with create_test_client(route_handlers=handler, experimental_features=experimental_features) as client:
        response = client.post("/", json={"what": "ever"})
        assert response.status_code == 201
        assert response.json() == {"a": 1, "b": "2"}


def test_dto_defined_on_controller(
    ModelDataDTO: type[AbstractDTO], experimental_features: list[ExperimentalFeatures]
) -> None:
    class MyController(Controller):
        dto = ModelDataDTO

        @post()
        def handler(self, data: Model) -> Model:
            assert data == Model(a=1, b="2")
            return data

    with create_test_client(route_handlers=MyController, experimental_features=experimental_features) as client:
        response = client.post("/", json={"what": "ever"})
        assert response.status_code == 201
        assert response.json() == {"a": 1, "b": "2"}


def test_dto_defined_on_router(
    ModelDataDTO: type[AbstractDTO], experimental_features: list[ExperimentalFeatures]
) -> None:
    @post()
    def handler(data: Model) -> Model:
        assert data == Model(a=1, b="2")
        return data

    router = Router(path="/", route_handlers=[handler], dto=ModelDataDTO)

    with create_test_client(route_handlers=router, experimental_features=experimental_features) as client:
        response = client.post("/", json={"what": "ever"})
        assert response.status_code == 201
        assert response.json() == {"a": 1, "b": "2"}


def test_dto_defined_on_app(ModelDataDTO: type[AbstractDTO], experimental_features: list[ExperimentalFeatures]) -> None:
    @post()
    def handler(data: Model) -> Model:
        assert data == Model(a=1, b="2")
        return data

    with create_test_client(
        route_handlers=handler, dto=ModelDataDTO, experimental_features=experimental_features
    ) as client:
        response = client.post("/", json={"what": "ever"})
        assert response.status_code == 201
        assert response.json() == {"a": 1, "b": "2"}


def test_set_dto_none_disables_inherited_dto(
    ModelDataDTO: type[AbstractDTO], experimental_features: list[ExperimentalFeatures]
) -> None:
    @post(dto=None, signature_namespace={"dict": Dict})
    def handler(data: dict[str, str]) -> dict[str, str]:
        assert data == {"hello": "world"}
        return data

    mock_dto = MagicMock(spec=ModelDataDTO)

    with create_test_client(
        route_handlers=handler,
        dto=mock_dto,  # pyright:ignore
        experimental_features=experimental_features,
    ) as client:
        response = client.post("/", json={"hello": "world"})
        assert response.status_code == 201
        assert response.json() == {"hello": "world"}
        mock_dto.assert_not_called()


def test_dto_and_return_dto(
    ModelDataDTO: type[AbstractDTO],
    ModelReturnDTO: type[AbstractDTO],
    experimental_features: list[ExperimentalFeatures],
) -> None:
    @post()
    def handler(data: Model) -> Model:
        assert data == Model(a=1, b="2")
        return data

    with create_test_client(
        route_handlers=handler, dto=ModelDataDTO, return_dto=ModelReturnDTO, experimental_features=experimental_features
    ) as client:
        response = client.post("/", json={"what": "ever"})
        assert response.status_code == 201
        assert response.json() == {"a": 1, "b": "2"}


def test_enable_experimental_backend(ModelDataDTO: type[AbstractDTO], use_experimental_dto_backend: bool) -> None:
    @post(dto=ModelDataDTO, signature_types=[Model])
    def handler(data: Model) -> Model:
        return data

    Litestar(
        route_handlers=[handler],
        experimental_features=[ExperimentalFeatures.DTO_CODEGEN] if use_experimental_dto_backend else None,
    )

    backend = handler.resolve_data_dto()._dto_backends[handler.handler_id]["data_backend"]  # type: ignore[union-attr]
    if use_experimental_dto_backend:
        assert isinstance(backend, DTOCodegenBackend)
    else:
        assert isinstance(backend, DTOBackend)


def test_enable_experimental_backend_override_in_dto_config(ModelDataDTO: type[AbstractDTO]) -> None:
    ModelDataDTO.config = DTOConfig(experimental_codegen_backend=False)

    @post(dto=ModelDataDTO, signature_types=[Model])
    def handler(data: Model) -> Model:
        return data

    Litestar(
        route_handlers=[handler],
        experimental_features=[ExperimentalFeatures.DTO_CODEGEN],
    )

    backend = handler.resolve_data_dto()._dto_backends[handler.handler_id]["data_backend"]  # type: ignore[union-attr]
    assert isinstance(backend, DTOBackend)


def test_use_codegen_backend_by_default(ModelDataDTO: type[AbstractDTO]) -> None:
    ModelDataDTO.config = DTOConfig()

    @post(dto=ModelDataDTO, signature_types=[Model])
    def handler(data: Model) -> Model:
        return data

    Litestar(route_handlers=[handler])

    backend = handler.resolve_data_dto()._dto_backends[handler.handler_id]["data_backend"]  # type: ignore[union-attr]
    assert isinstance(backend, DTOBackend)



import subprocess
import time
from pathlib import Path
from typing import Callable, List

import httpx
import psutil
import pytest
from _pytest.fixtures import FixtureRequest
from _pytest.monkeypatch import MonkeyPatch


@pytest.fixture()
def run_server(tmp_path: Path, request: FixtureRequest, monkeypatch: MonkeyPatch) -> Callable[[str, List[str]], None]:
    def runner(app: str, server_command: List[str]) -> None:
        tmp_path.joinpath("app.py").write_text(app)
        monkeypatch.chdir(tmp_path)

        proc = psutil.Popen(
            server_command,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )

        def kill() -> None:
            for child in proc.children(recursive=True):
                child.kill()
            proc.kill()

        request.addfinalizer(kill)

        for _ in range(50):
            try:
                httpx.get("http://127.0.0.1:9999/", timeout=0.1)
                break
            except httpx.TransportError:
                time.sleep(0.1)
        else:
            raise RuntimeError("App failed to come online")

    return runner



from dataclasses import dataclass
from typing import List, Optional

import pytest
from attr import define
from typing_extensions import Annotated, TypedDict

from litestar import get, post
from litestar._signature import SignatureModel
from litestar.di import Provide
from litestar.enums import ParamType
from litestar.exceptions import ImproperlyConfiguredException, ValidationException
from litestar.params import Dependency, Parameter
from litestar.status_codes import HTTP_400_BAD_REQUEST, HTTP_500_INTERNAL_SERVER_ERROR
from litestar.testing import RequestFactory, create_test_client
from litestar.utils.signature import ParsedSignature


def test_parses_values_from_connection_kwargs_raises() -> None:
    def fn(a: int) -> None:
        pass

    model = SignatureModel.create(
        dependency_name_set=set(),
        fn=fn,
        data_dto=None,
        parsed_signature=ParsedSignature.from_fn(fn, {}),
        type_decoders=[],
    )
    with pytest.raises(ValidationException):
        model.parse_values_from_connection_kwargs(connection=RequestFactory().get(), a="not an int")


def test_create_signature_validation() -> None:
    @get()
    def my_fn(typed: int, untyped) -> None:  # type: ignore[no-untyped-def]
        pass

    with pytest.raises(ImproperlyConfiguredException):
        SignatureModel.create(
            dependency_name_set=set(),
            fn=my_fn.fn,
            data_dto=None,
            parsed_signature=ParsedSignature.from_fn(my_fn.fn, {}),
            type_decoders=[],
        )


def test_dependency_validation_failure_raises_500() -> None:
    dependencies = {"dep": Provide(lambda: "thirteen", sync_to_thread=False)}

    @get("/")
    def test(dep: int, param: int, optional_dep: Optional[int] = Dependency()) -> None: ...

    with create_test_client(
        route_handlers=[test],
        dependencies=dependencies,
        debug=False,
    ) as client:
        response = client.get("/?param=13")

    assert response.json() == {"detail": "Internal Server Error", "status_code": HTTP_500_INTERNAL_SERVER_ERROR}


def test_validation_failure_raises_400() -> None:
    dependencies = {"dep": Provide(lambda: 13, sync_to_thread=False)}

    @get("/")
    def test(dep: int, param: int, optional_dep: Optional[int] = Dependency()) -> None: ...

    with create_test_client(route_handlers=[test], dependencies=dependencies) as client:
        response = client.get("/?param=thirteen")

    assert response.json() == {
        "detail": "Validation failed for GET /?param=thirteen",
        "extra": [{"key": "param", "message": "Expected `int`, got `str`", "source": "query"}],
        "status_code": 400,
    }


def test_client_backend_error_precedence_over_server_error() -> None:
    dependencies = {
        "dep": Provide(lambda: "thirteen", sync_to_thread=False),
        "optional_dep": Provide(lambda: "thirty-one", sync_to_thread=False),
    }

    @get("/")
    def test(dep: int, param: int, optional_dep: Optional[int] = Dependency()) -> None: ...

    with create_test_client(route_handlers=[test], dependencies=dependencies) as client:
        response = client.get("/?param=thirteen")

    assert response.json() == {
        "detail": "Validation failed for GET /?param=thirteen",
        "extra": [{"key": "param", "message": "Expected `int`, got `str`", "source": "query"}],
        "status_code": 400,
    }


def test_validation_error_exception_key() -> None:
    from dataclasses import dataclass

    @dataclass
    class OtherChild:
        val: List[int]

    @dataclass
    class Child:
        val: int
        other_val: int

    @dataclass
    class Parent:
        child: Child
        other_child: OtherChild

    @get("/")
    def handler(data: Parent) -> None:
        pass

    model = SignatureModel.create(
        dependency_name_set=set(),
        fn=handler,
        data_dto=None,
        parsed_signature=ParsedSignature.from_fn(handler.fn, {}),
        type_decoders=[],
    )

    with pytest.raises(ValidationException) as exc_info:
        model.parse_values_from_connection_kwargs(
            connection=RequestFactory().get(route_handler=handler), data={"child": {}, "other_child": {}}
        )

    assert isinstance(exc_info.value.extra, list)
    assert exc_info.value.extra[0]["key"] == "child"


def test_invalid_input_attrs() -> None:
    @define
    class OtherChild:
        val: List[int]

    @define
    class Child:
        val: int
        other_val: int

    @define
    class Parent:
        child: Child
        other_child: OtherChild

    @post("/")
    def test(
        data: Parent,
        int_param: int,
        int_header: int = Parameter(header="X-SOME-INT"),
        int_cookie: int = Parameter(cookie="int-cookie"),
    ) -> None: ...

    with create_test_client(route_handlers=[test]) as client:
        client.cookies.update({"int-cookie": "cookie"})
        response = client.post(
            "/",
            json={"child": {"val": "a", "other_val": "b"}, "other_child": {"val": [1, "c"]}},
            params={"int_param": "param"},
            headers={"X-SOME-INT": "header"},
        )

        assert response.status_code == HTTP_400_BAD_REQUEST

        data = response.json()

        assert data
        assert data["extra"] == [
            {"message": "Expected `int`, got `str`", "key": "child.val", "source": "body"},
            {"message": "Expected `int`, got `str`", "key": "int_param", "source": "query"},
            {"message": "Expected `int`, got `str`", "key": "int_header", "source": "header"},
            {"message": "Expected `int`, got `str`", "key": "int_cookie", "source": "cookie"},
        ]


def test_invalid_input_dataclass() -> None:
    @dataclass
    class OtherChild:
        val: List[int]

    @dataclass
    class Child:
        val: int
        other_val: int

    @dataclass
    class Parent:
        child: Child
        other_child: OtherChild

    @post("/")
    def test(
        data: Parent,
        int_param: int,
        length_param: str = Parameter(min_length=2),
        int_header: int = Parameter(header="X-SOME-INT"),
        int_cookie: int = Parameter(cookie="int-cookie"),
    ) -> None: ...

    with create_test_client(route_handlers=[test]) as client:
        client.cookies.update({"int-cookie": "cookie"})
        response = client.post(
            "/",
            json={"child": {"val": "a", "other_val": "b"}, "other_child": {"val": [1, "c"]}},
            params={"int_param": "param", "length_param": "d"},
            headers={"X-SOME-INT": "header"},
        )

        assert response.status_code == HTTP_400_BAD_REQUEST

        data = response.json()

        assert data
        assert data["extra"] == [
            {"message": "Expected `int`, got `str`", "key": "child.val", "source": "body"},
            {"message": "Expected `int`, got `str`", "key": "int_param", "source": "query"},
            {"message": "Expected `str` of length >= 2", "key": "length_param", "source": "query"},
            {"message": "Expected `int`, got `str`", "key": "int_header", "source": "header"},
            {"message": "Expected `int`, got `str`", "key": "int_cookie", "source": "cookie"},
        ]


def test_invalid_input_typed_dict() -> None:
    class OtherChild(TypedDict):
        val: List[int]

    class Child(TypedDict):
        val: int
        other_val: int

    class Parent(TypedDict):
        child: Child
        other_child: OtherChild

    @post("/")
    def test(
        data: Parent,
        int_param: int,
        length_param: str = Parameter(min_length=2),
        int_header: int = Parameter(header="X-SOME-INT"),
        int_cookie: int = Parameter(cookie="int-cookie"),
    ) -> None: ...

    with create_test_client(route_handlers=[test]) as client:
        client.cookies.update({"int-cookie": "cookie"})
        response = client.post(
            "/",
            json={"child": {"val": "a", "other_val": "b"}, "other_child": {"val": [1, "c"]}},
            params={"int_param": "param", "length_param": "d"},
            headers={"X-SOME-INT": "header"},
        )

        assert response.status_code == HTTP_400_BAD_REQUEST

        data = response.json()

        assert data
        assert data["extra"] == [
            {"message": "Expected `int`, got `str`", "key": "child.val", "source": "body"},
            {"message": "Expected `int`, got `str`", "key": "int_param", "source": "query"},
            {"message": "Expected `str` of length >= 2", "key": "length_param", "source": "query"},
            {"message": "Expected `int`, got `str`", "key": "int_header", "source": "header"},
            {"message": "Expected `int`, got `str`", "key": "int_cookie", "source": "cookie"},
        ]


def test_parse_values_from_connection_kwargs_with_multiple_errors() -> None:
    def fn(a: Annotated[int, Parameter(gt=5)], b: Annotated[int, Parameter(lt=5)]) -> None:
        pass

    model = SignatureModel.create(
        dependency_name_set=set(),
        fn=fn,
        data_dto=None,
        parsed_signature=ParsedSignature.from_fn(fn, {}),
        type_decoders=[],
    )
    with pytest.raises(ValidationException) as exc:
        model.parse_values_from_connection_kwargs(connection=RequestFactory().get(), a=0, b=9)

    assert exc.value.extra == [
        {"message": "Expected `int` >= 6", "key": "a", "source": ParamType.QUERY},
        {"message": "Expected `int` <= 4", "key": "b", "source": ParamType.QUERY},
    ]



import os
import secrets
import time
from base64 import b64decode, b64encode
from typing import Any, Dict
from unittest import mock

import pytest
from cryptography.exceptions import InvalidTag

from litestar import Request, get, post
from litestar.exceptions import ImproperlyConfiguredException
from litestar.middleware.session import SessionMiddleware
from litestar.middleware.session.client_side import (
    AAD,
    CHUNK_SIZE,
    ClientSideSessionBackend,
    CookieBackendConfig,
)
from litestar.serialization import encode_json
from litestar.testing import create_test_client
from tests.helpers import randbytes


@pytest.mark.parametrize(
    "secret, should_raise",
    [
        [randbytes(16), False],
        [randbytes(24), False],
        [randbytes(32), False],
        [randbytes(17), True],
        [randbytes(4), True],
        [randbytes(100), True],
        [b"", True],
    ],
)
def test_secret_validation(secret: bytes, should_raise: bool) -> None:
    if should_raise:
        with pytest.raises(ImproperlyConfiguredException):
            CookieBackendConfig(secret=secret)
    else:
        CookieBackendConfig(secret=secret)


@pytest.mark.parametrize(
    "key, should_raise",
    [
        ["", True],
        ["a", False],
        ["a" * 256, False],
        ["a" * 257, True],
    ],
)
def test_key_validation(key: str, should_raise: bool) -> None:
    if should_raise:
        with pytest.raises(ImproperlyConfiguredException):
            CookieBackendConfig(secret=os.urandom(16), key=key)
    else:
        CookieBackendConfig(secret=os.urandom(16), key=key)


@pytest.mark.parametrize(
    "max_age, should_raise",
    [
        [0, True],
        [-1, True],
        [1, False],
        [100, False],
    ],
)
def test_max_age_validation(max_age: int, should_raise: bool) -> None:
    if should_raise:
        with pytest.raises(ImproperlyConfiguredException):
            CookieBackendConfig(secret=os.urandom(16), key="a", max_age=max_age)
    else:
        CookieBackendConfig(secret=os.urandom(16), key="a", max_age=max_age)


def create_session(size: int = 16) -> Dict[str, str]:
    return {"key": secrets.token_hex(size)}


@pytest.mark.parametrize("session", [create_session(), create_session(size=4096)])
def test_dump_and_load_data(session: dict, cookie_session_backend: ClientSideSessionBackend) -> None:
    ciphertext = cookie_session_backend.dump_data(session)
    assert isinstance(ciphertext, list)

    for text in ciphertext:
        assert len(text) <= CHUNK_SIZE

    plain_text = cookie_session_backend.load_data(ciphertext)
    assert plain_text == session


@mock.patch("time.time", return_value=round(time.time()))
def test_load_data_should_return_empty_if_session_expired(
    time_mock: mock.MagicMock, cookie_session_backend: ClientSideSessionBackend
) -> None:
    """Should return empty dict if session is expired."""
    ciphertext = cookie_session_backend.dump_data(create_session())
    time_mock.return_value = round(time.time()) + cookie_session_backend.config.max_age + 1
    plaintext = cookie_session_backend.load_data(data=ciphertext)
    assert plaintext == {}


def test_set_session_cookies(cookie_session_backend_config: "CookieBackendConfig") -> None:
    """Should set session cookies from session in response."""
    chunks_multiplier = 2

    @get(path="/test")
    def handler(request: Request) -> None:
        # Create large session by keeping it multiple of CHUNK_SIZE. This will split the session into multiple cookies.
        # Then you only need to check if number of cookies set are more than the multiplying number.
        request.session.update(create_session(size=CHUNK_SIZE * chunks_multiplier))

    @get(path="/test_short_cookie")
    def handler_short_cookie(request: Request) -> None:
        # Check the naming of a cookie that's short enough to not get broken into chunks
        request.session.update(create_session())

    with create_test_client(
        route_handlers=[handler],
        middleware=[cookie_session_backend_config.middleware],
    ) as client:
        response = client.get("/test")

        assert len(response.cookies) > chunks_multiplier
        assert "session-0" in response.cookies

    with create_test_client(
        route_handlers=[handler_short_cookie],
        middleware=[cookie_session_backend_config.middleware],
    ) as client:
        response = client.get("/test_short_cookie")

        assert len(response.cookies) == 1
        assert "session" in response.cookies


def test_session_cookie_name_matching(cookie_session_backend_config: "CookieBackendConfig") -> None:
    session_data = {"foo": "bar"}

    @get("/")
    def handler(request: Request) -> Dict[str, Any]:
        return request.session

    @post("/")
    def set_session_data(request: Request) -> None:
        request.set_session(session_data)

    with create_test_client(
        route_handlers=[handler, set_session_data],
        middleware=[cookie_session_backend_config.middleware],
    ) as client:
        client.post("/")
        client.cookies[f"thisisnnota{cookie_session_backend_config.key}cookie"] = "foo"
        response = client.get("/")
        assert response.json() == session_data


@pytest.mark.parametrize("mutate", [False, True])
def test_load_session_cookies_and_expire_previous(
    mutate: bool, cookie_session_middleware: SessionMiddleware[ClientSideSessionBackend]
) -> None:
    """Should load session cookies into session from request and overwrite the previously set cookies with the upcoming
    response.

    Session cookies from the previous session should not persist because session is mutable. Once the session is loaded
    from the cookies, those cookies are redundant. The response sets new session cookies overwriting or expiring the
    previous ones.
    """
    # Test for large session data. If it works for multiple cookies, it works for single also.
    _session = create_session(size=4096)

    @get(path="/test")
    def handler(request: Request) -> dict:
        nonlocal _session
        if mutate:
            # Modify the session, this will overwrite the previously set session cookies.
            request.session.update(create_session())
            _session = request.session
        return request.session

    ciphertext = cookie_session_middleware.backend.dump_data(_session)

    with create_test_client(
        route_handlers=[handler],
        middleware=[cookie_session_middleware.backend.config.middleware],
    ) as client:
        # Set cookies on the client to avoid warnings about per-request cookies.
        client.cookies = {  # type: ignore[assignment]
            f"{cookie_session_middleware.backend.config.key}-{i}": text.decode("utf-8")
            for i, text in enumerate(ciphertext)
        }
        response = client.get("/test")

    assert response.json() == _session
    # The session cookie names that were in the request will also be present in its response to overwrite or to expire
    # them. So, the number of cookies in the response will be at least equal to or greater than the number of cookies
    # that were in the request.
    assert response.headers["set-cookie"].count("session") >= response.request.headers["Cookie"].count("session")


def test_load_data_should_raise_invalid_tag_if_tampered_aad(cookie_session_backend: ClientSideSessionBackend) -> None:
    """If AAD has been tampered with, the integrity of the data cannot be verified and InavlidTag exception is
    raised.
    """
    encrypted_session = cookie_session_backend.dump_data(create_session())
    # The attacker will tamper with the AAD to increase the expiry time of the cookie.
    attacker_chosen_time = 300  # In seconds
    fraudulent_associated_data = encode_json(
        {"expires_at": round(time.time()) + cookie_session_backend.config.max_age + attacker_chosen_time}
    )
    decoded = b64decode(b"".join(encrypted_session))
    aad_starts_from = decoded.find(AAD)
    # The attacker removes the original AAD and attaches its own.
    ciphertext = b64encode(decoded[:aad_starts_from] + AAD + fraudulent_associated_data)
    # The attacker puts the data back to its original form.
    encoded = [ciphertext[i : i + CHUNK_SIZE] for i in range(0, len(ciphertext), CHUNK_SIZE)]

    with pytest.raises(InvalidTag):
        cookie_session_backend.load_data(encoded)



import pytest

from litestar.app import Litestar
from litestar.repository import handlers
from litestar.repository.filters import (
    BeforeAfter,
    CollectionFilter,
    FilterTypes,
    LimitOffset,
    NotInCollectionFilter,
    NotInSearchFilter,
    OnBeforeAfter,
    OrderBy,
    SearchFilter,
)


def test_app_repository_signature_namespace() -> None:
    app = Litestar([], on_app_init=[handlers.on_app_init])

    assert app.signature_namespace == {
        "BeforeAfter": BeforeAfter,
        "OnBeforeAfter": OnBeforeAfter,
        "CollectionFilter": CollectionFilter,
        "LimitOffset": LimitOffset,
        "OrderBy": OrderBy,
        "SearchFilter": SearchFilter,
        "NotInCollectionFilter": NotInCollectionFilter,
        "NotInSearchFilter": NotInSearchFilter,
        "FilterTypes": FilterTypes,
    }


def test_deprecated_abc_imports() -> None:
    from litestar.contrib.repository import abc as abc_contrib
    from litestar.repository import abc

    assert abc_contrib.AbstractAsyncRepository is abc.AbstractAsyncRepository
    assert abc_contrib.AbstractSyncRepository is abc.AbstractSyncRepository

    with pytest.raises(AttributeError):
        abc_contrib.foo


def test_deprecated_exception_imports() -> None:
    from litestar.contrib.repository import exceptions as contrib_exceptions
    from litestar.repository import exceptions

    assert exceptions.RepositoryError is contrib_exceptions.RepositoryError
    assert exceptions.ConflictError is contrib_exceptions.ConflictError
    assert exceptions.NotFoundError is contrib_exceptions.NotFoundError

    with pytest.raises(AttributeError):
        contrib_exceptions.foo


def test_deprecated_filters_imports() -> None:
    from litestar.contrib.repository import filters as contrib_filter
    from litestar.repository import filters

    assert filters.FilterTypes is contrib_filter.FilterTypes
    assert filters.CollectionFilter is contrib_filter.CollectionFilter
    assert filters.NotInCollectionFilter is contrib_filter.NotInCollectionFilter
    assert filters.SearchFilter is contrib_filter.SearchFilter
    assert filters.NotInSearchFilter is contrib_filter.NotInSearchFilter
    assert filters.OnBeforeAfter is contrib_filter.OnBeforeAfter
    assert filters.BeforeAfter is contrib_filter.BeforeAfter
    assert filters.LimitOffset is contrib_filter.LimitOffset
    assert filters.OrderBy is contrib_filter.OrderBy

    with pytest.raises(AttributeError):
        contrib_filter.foo


def test_advanced_alchemy_imports() -> None:
    from advanced_alchemy import filters

    from litestar.repository import _filters

    assert filters.FilterTypes is not _filters.FilterTypes
    assert filters.CollectionFilter is not _filters.CollectionFilter
    assert filters.NotInCollectionFilter is not _filters.NotInCollectionFilter
    assert filters.SearchFilter is not _filters.SearchFilter
    assert filters.NotInSearchFilter is not _filters.NotInSearchFilter
    assert filters.OnBeforeAfter is not _filters.OnBeforeAfter
    assert filters.BeforeAfter is not _filters.BeforeAfter
    assert filters.LimitOffset is not _filters.LimitOffset
    assert filters.OrderBy is not _filters.OrderBy


def test_deprecated_handlers_imports() -> None:
    from litestar.contrib.repository import handlers as contrib_handlers
    from litestar.repository import handlers

    assert handlers.on_app_init is contrib_handlers.on_app_init

    with pytest.raises(AttributeError):
        contrib_handlers.foo


def test_deprecated_testing_imports() -> None:
    from litestar.contrib.repository import testing as contrib_testing
    from litestar.repository.testing import generic_mock_repository

    assert generic_mock_repository.GenericAsyncMockRepository is contrib_testing.GenericAsyncMockRepository
    assert generic_mock_repository.GenericSyncMockRepository is contrib_testing.GenericSyncMockRepository

    with pytest.raises(AttributeError):
        contrib_testing.foo



from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, List, Optional, Union

import pytest
from pytest import FixtureRequest

from litestar import MediaType
from litestar.datastructures import (
    Accept,
    CacheControlHeader,
    ETag,
    Headers,
    MutableScopeHeaders,
)
from litestar.datastructures.headers import Header
from litestar.exceptions import ImproperlyConfiguredException, ValidationException
from litestar.types.asgi_types import HTTPResponseBodyEvent, HTTPResponseStartEvent
from litestar.utils.dataclass import simple_asdict

if TYPE_CHECKING:
    from litestar.types.asgi_types import RawHeaders, RawHeadersList, Scope


@pytest.fixture
def raw_headers() -> "RawHeadersList":
    return [(b"foo", b"bar")]


@pytest.fixture
def raw_headers_tuple() -> "RawHeaders":
    return [(b"foo", b"bar")]


def test_header_container_requires_header_key_being_defined() -> None:
    class TestHeader(Header):
        def _get_header_value(self) -> str:
            return ""

        def from_header(self, header_value: str) -> "Header":  # type: ignore[explicit-override, override]
            return self

    with pytest.raises(ImproperlyConfiguredException):
        TestHeader().to_header()


@pytest.fixture
def mutable_headers(raw_headers: "RawHeadersList") -> MutableScopeHeaders:
    return MutableScopeHeaders({"headers": raw_headers})


@pytest.fixture
def mutable_headers_from_tuple(raw_headers_tuple: "RawHeaders") -> MutableScopeHeaders:
    return MutableScopeHeaders({"headers": raw_headers_tuple})


@pytest.fixture(params=[True, False])
def existing_headers_key(request: FixtureRequest) -> str:
    return "Foo" if request.param else "foo"


def test_headers_from_mapping() -> None:
    headers = Headers({"foo": "bar", "baz": "zab"})
    assert headers["foo"] == "bar"
    assert headers["baz"] == "zab"


def test_headers_from_raw_list() -> None:
    headers = Headers([(b"foo", b"bar"), (b"foo", b"baz")])
    assert headers.getall("foo") == ["bar", "baz"]


def test_headers_from_raw_tuple() -> None:
    headers = Headers(((b"foo", b"bar"), (b"foo", b"baz")))
    assert headers.getall("foo") == ["bar", "baz"]


def test_headers_from_scope(create_scope: "Callable[..., Scope]") -> None:
    headers = Headers.from_scope(create_scope(headers=[(b"foo", b"bar"), (b"buzz", b"bup")]))
    assert headers["foo"] == "bar"
    assert headers["buzz"] == "bup"


def test_headers_to_header_list() -> None:
    raw = [(b"foo", b"bar"), (b"foo", b"baz")]
    headers = Headers(raw)
    assert headers.to_header_list() == raw


def test_headers_tuple_to_header_list() -> None:
    raw = ((b"foo", b"bar"), (b"foo", b"baz"))
    headers = Headers(raw)
    assert headers.to_header_list() == list(raw)


def test_mutable_scope_headers_from_iterable() -> None:
    raw = [(b"foo", b"bar"), (b"foo", b"baz")]
    headers = MutableScopeHeaders({"headers": iter(raw)})
    assert headers.headers == raw


def test_mutable_scope_headers_from_message(raw_headers: "RawHeadersList", raw_headers_tuple: "RawHeaders") -> None:
    headers = MutableScopeHeaders.from_message(
        HTTPResponseStartEvent(type="http.response.start", status=200, headers=raw_headers)
    )
    assert headers.headers == raw_headers

    headers = MutableScopeHeaders.from_message(
        HTTPResponseStartEvent(type="http.response.start", status=200, headers=raw_headers_tuple)
    )
    assert headers.headers == list(raw_headers_tuple)


def test_mutable_scope_headers_from_message_invalid_type() -> None:
    with pytest.raises(ValueError):
        MutableScopeHeaders.from_message(HTTPResponseBodyEvent(type="http.response.body", body=b"", more_body=False))


def test_mutable_scope_headers_add(
    raw_headers: "RawHeadersList", mutable_headers: MutableScopeHeaders, existing_headers_key: str
) -> None:
    mutable_headers.add(existing_headers_key, "baz")
    assert raw_headers == [(b"foo", b"bar"), (b"foo", b"baz")]


def test_mutable_scope_headers_from_tuple_add(
    raw_headers_tuple: "RawHeaders", mutable_headers_from_tuple: MutableScopeHeaders, existing_headers_key: str
) -> None:
    mutable_headers_from_tuple.add(existing_headers_key, "baz")
    assert list(raw_headers_tuple) == [(b"foo", b"bar"), (b"foo", b"baz")]


def test_mutable_scope_headers_getall_singular_value(
    mutable_headers: MutableScopeHeaders, mutable_headers_from_tuple: MutableScopeHeaders, existing_headers_key: str
) -> None:
    assert mutable_headers.getall(existing_headers_key) == ["bar"]
    assert mutable_headers_from_tuple.getall(existing_headers_key) == ["bar"]


def test_mutable_scope_headers_getall_multi_value(
    mutable_headers: MutableScopeHeaders, mutable_headers_from_tuple: MutableScopeHeaders, existing_headers_key: str
) -> None:
    mutable_headers.add(existing_headers_key, "baz")
    assert mutable_headers.getall("foo") == ["bar", "baz"]

    mutable_headers_from_tuple.add(existing_headers_key, "baz")
    assert mutable_headers.getall("foo") == ["bar", "baz"]


def test_mutable_scope_headers_getall_not_found_no_default(mutable_headers: MutableScopeHeaders) -> None:
    with pytest.raises(KeyError):
        mutable_headers.getall("bar")


def test_mutable_scope_headers_getall_not_found_default(mutable_headers: MutableScopeHeaders) -> None:
    assert mutable_headers.getall("bar", ["default"]) == ["default"]


def test_mutable_scope_headers_extend_header_value(
    raw_headers: "RawHeadersList", mutable_headers: MutableScopeHeaders
) -> None:
    mutable_headers.extend_header_value("foo", "baz")
    assert raw_headers == [(b"foo", b"bar,baz")]


def test_mutable_scope_headers_from_tuple_extend_header_value(
    raw_headers_tuple: "RawHeaders", mutable_headers_from_tuple: MutableScopeHeaders
) -> None:
    mutable_headers_from_tuple.extend_header_value("foo", "baz")
    assert list(raw_headers_tuple) == [(b"foo", b"bar,baz")]


def test_mutable_scope_headers_extend_header_value_new_header(
    raw_headers: "RawHeadersList", mutable_headers: MutableScopeHeaders
) -> None:
    mutable_headers.extend_header_value("bar", "baz")
    assert raw_headers == [(b"foo", b"bar"), (b"bar", b"baz")]


def test_mutable_scope_headers_from_tuple_extend_header_value_new_header(
    raw_headers_tuple: "RawHeaders", mutable_headers_from_tuple: MutableScopeHeaders
) -> None:
    mutable_headers_from_tuple.extend_header_value("bar", "baz")
    assert list(raw_headers_tuple) == [(b"foo", b"bar"), (b"bar", b"baz")]


def test_mutable_scope_headers_getitem(mutable_headers: MutableScopeHeaders, existing_headers_key: str) -> None:
    assert mutable_headers[existing_headers_key] == "bar"


def test_mutable_scope_headers_getitem_not_found(mutable_headers: MutableScopeHeaders) -> None:
    with pytest.raises(KeyError):
        mutable_headers["bar"]


def test_mutable_scope_headers_setitem_existing_key(
    raw_headers: "RawHeadersList", mutable_headers: MutableScopeHeaders, existing_headers_key: str
) -> None:
    mutable_headers[existing_headers_key] = "baz"
    assert raw_headers == [(b"foo", b"baz")]


def test_mutable_scope_headers_from_tuple_setitem_existing_key(
    raw_headers_tuple: "RawHeaders", mutable_headers_from_tuple: MutableScopeHeaders, existing_headers_key: str
) -> None:
    mutable_headers_from_tuple[existing_headers_key] = "baz"
    assert list(raw_headers_tuple) == [(b"foo", b"baz")]


def test_mutable_scope_headers_setitem_new_key(
    raw_headers: "RawHeadersList", mutable_headers: MutableScopeHeaders
) -> None:
    mutable_headers["bar"] = "baz"
    assert raw_headers == [(b"foo", b"bar"), (b"bar", b"baz")]


def test_mutable_scope_headers_setitem_delitem(
    raw_headers: "RawHeadersList", mutable_headers: MutableScopeHeaders, existing_headers_key: str
) -> None:
    mutable_headers.add("foo", "baz")
    mutable_headers["bar"] = "baz"
    del mutable_headers[existing_headers_key]
    assert raw_headers == [(b"bar", b"baz")]


def test_mutable_scope_headers_setdefault() -> None:
    headers = MutableScopeHeaders()

    assert headers.setdefault("foo", "bar") == "bar"
    assert headers.setdefault("foo", "baz") == "bar"
    assert headers.getall("foo") == ["bar"]


def test_mutable_scope_header_len(mutable_headers: MutableScopeHeaders) -> None:
    assert len(mutable_headers) == 1
    mutable_headers.add("foo", "bar")
    assert len(mutable_headers) == 2
    mutable_headers["bar"] = "baz"
    assert len(mutable_headers) == 3


def test_mutable_scope_header_iter(mutable_headers: MutableScopeHeaders) -> None:
    mutable_headers.add("foo", "baz")
    mutable_headers["bar"] = "zab"
    assert list(mutable_headers) == ["foo", "foo", "bar"]


def test_cache_control_to_header() -> None:
    header = CacheControlHeader(max_age=10, private=True)
    expected_header_values = ["max-age=10, private", "private, max-age=10"]
    assert header.to_header() in expected_header_values
    assert header.to_header(include_header_name=True) in [f"cache-control: {v}" for v in expected_header_values]


def test_cache_control_from_header() -> None:
    header_value = (
        "public, private, no-store, no-cache, max-age=10000, s-maxage=1000, no-transform, "
        "must-revalidate, proxy-revalidate, must-understand, immutable, stale-while-revalidate=100"
    )
    header = CacheControlHeader.from_header(header_value)
    assert simple_asdict(header) == {
        "documentation_only": False,
        "public": True,
        "private": True,
        "no_store": True,
        "no_cache": True,
        "max_age": 10000,
        "s_maxage": 1000,
        "no_transform": True,
        "must_revalidate": True,
        "proxy_revalidate": True,
        "must_understand": True,
        "immutable": True,
        "stale_while_revalidate": 100,
    }


def test_cache_control_from_header_single_value() -> None:
    header_value = "no-cache"
    header = CacheControlHeader.from_header(header_value)
    header_dict = simple_asdict(header, exclude_none=True)
    assert header_dict == {"no_cache": True, "documentation_only": False}


@pytest.mark.parametrize("invalid_value", ["x=y=z", "x, ", "no-cache=10", "invalid-header=10"])
def test_cache_control_from_header_invalid_value(invalid_value: str) -> None:
    with pytest.raises(ImproperlyConfiguredException):
        CacheControlHeader.from_header(invalid_value)


def test_cache_control_header_prevent_storing() -> None:
    header = CacheControlHeader.prevent_storing()
    header_dict = simple_asdict(header, exclude_none=True)
    assert header_dict == {"no_store": True, "documentation_only": False}


def test_cache_control_header_unsupported_type_annotation() -> None:
    @dataclass
    class InvalidCacheControlHeader(CacheControlHeader):
        unsupported_type: Union[int, str] = "foo"

    with pytest.raises(ImproperlyConfiguredException):
        InvalidCacheControlHeader.from_header("unsupported_type")


def test_etag_documentation_only() -> None:
    assert ETag(documentation_only=True).value is None


def test_etag_no_value() -> None:
    with pytest.raises(ValidationException):
        ETag()

    with pytest.raises(ValidationException):
        ETag(weak=True)


def test_etag_non_ascii() -> None:
    with pytest.raises(ValueError):
        ETag(value="f↓o")


def test_etag_from_header() -> None:
    etag = ETag.from_header('"foo"')
    assert etag.value == "foo"
    assert etag.weak is False


@pytest.mark.parametrize("value", ['W/"foo"', 'w/"foo"'])
def test_etag_from_header_weak(value: str) -> None:
    etag = ETag.from_header(value)
    assert etag.value == "foo"
    assert etag.weak is True


@pytest.mark.parametrize("value", ['"føo"', 'W/"føo"'])
def test_etag_from_header_non_ascii_value(value: str) -> None:
    with pytest.raises(ImproperlyConfiguredException):
        ETag.from_header(value)


@pytest.mark.parametrize("value", ["foo", "W/foo"])
def test_etag_from_header_missing_quotes(value: str) -> None:
    with pytest.raises(ImproperlyConfiguredException):
        ETag.from_header(value)


def test_etag_to_header() -> None:
    assert ETag(value="foo").to_header() == '"foo"'


def test_etag_to_header_weak() -> None:
    assert ETag(value="foo", weak=True).to_header() == 'W/"foo"'


@pytest.mark.parametrize(
    "accept_value,provided_types,best_match",
    (
        ("text/plain", ["text/plain"], "text/plain"),
        ("text/plain", [MediaType.TEXT], MediaType.TEXT),
        ("text/plain", ["text/plain"], "text/plain"),
        ("text/plain", ["text/html"], None),
        ("text/*", ["text/html"], "text/html"),
        ("*/*", ["text/html"], "text/html"),
        ("text/plain;p=test", ["text/plain"], "text/plain"),
        ("text/plain", ["text/plain;p=test"], None),
        ("text/plain;p=test", ["text/plain;p=test"], "text/plain;p=test"),
        ("text/plain", ["text/*"], "text/plain"),
        ("text/html", ["*/*"], "text/html"),
        ("text/plain;q=0.8,text/html", ["text/plain", "text/html"], "text/html"),
        ("text/*,text/html", ["text/plain", "text/html"], "text/html"),
    ),
)
def test_accept_best_match(accept_value: str, provided_types: List[str], best_match: Optional[str]) -> None:
    accept = Accept(accept_value)
    assert accept.best_match(provided_types) == best_match


def test_accept_accepts() -> None:
    accept = Accept("text/plain;q=0.8,text/html")
    assert accept.accepts(MediaType.TEXT)



from __future__ import annotations

import inspect
from inspect import Signature
from typing import Any

from litestar.contrib.pydantic.utils import is_pydantic_model_class
from litestar.plugins import DIPlugin


class PydanticDIPlugin(DIPlugin):
    def has_typed_init(self, type_: Any) -> bool:
        return is_pydantic_model_class(type_)

    def get_typed_init(self, type_: Any) -> tuple[Signature, dict[str, Any]]:
        try:
            model_fields = dict(type_.model_fields)
        except AttributeError:
            model_fields = {k: model_field.field_info for k, model_field in type_.__fields__.items()}

        parameters = [
            inspect.Parameter(name=field_name, kind=inspect.Parameter.KEYWORD_ONLY, annotation=Any)
            for field_name in model_fields
        ]
        type_hints = {field_name: Any for field_name in model_fields}
        return Signature(parameters), type_hints



import secrets
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from litestar.middleware.session.client_side import ClientSideSessionBackend


@pytest.fixture()
def session_test_cookies(cookie_session_backend: "ClientSideSessionBackend") -> str:
    # Put random data. If you are also handling session management then use session_middleware fixture and create
    # session cookies with your own data.
    _session = {"key": secrets.token_hex(16)}
    return "; ".join(
        f"session-{i}={serialize.decode('utf-8')}"
        for i, serialize in enumerate(cookie_session_backend.dump_data(_session))
    )



from __future__ import annotations

from dataclasses import dataclass

from litestar.openapi.spec.base import BaseSchemaObject

__all__ = ("XML",)


@dataclass()
class XML(BaseSchemaObject):
    """A metadata object that allows for more fine-tuned XML model definitions.

    When using arrays, XML element names are *not* inferred (for singular/plural forms) and the ``name`` property SHOULD
    be used to add that information. See examples for expected behavior.
    """

    name: str | None = None
    """
    Replaces the name of the element/attribute used for the described schema property. When defined within ``items``, it
    will affect the name of the individual XML elements within the list. When defined alongside ``type`` being ``array``
    (outside the ``items``), it will affect the wrapping element and only if ``wrapped`` is ``True``. If ``wrapped`` is
    ``False``, it will be ignored.
    """

    namespace: str | None = None
    """The URI of the namespace definition. Value MUST be in the form of an absolute URI."""

    prefix: str | None = None
    """The prefix to be used for the
    `xmlName <https://spec.openapis.org/oas/v3.1.0#xmlName>`_
    """

    attribute: bool = False
    """Declares whether the property definition translates to an attribute instead of an element. Default value is
    ``False``.
    """

    wrapped: bool = False
    """
    MAY be used only for an array definition. Signifies whether the array is wrapped (for example,
    ``<books><book/><book/></books>``) or unwrapped (``<book/><book/>``). Default value is ``False``. The definition
    takes effect only when defined alongside ``type`` being ``array`` (outside the ``items``).
    """



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from litestar.cli._utils import console
from litestar.logging.config import StructLoggingConfig
from litestar.middleware.logging import LoggingMiddlewareConfig
from litestar.plugins import InitPluginProtocol

if TYPE_CHECKING:
    from litestar.config.app import AppConfig


@dataclass
class StructlogConfig:
    structlog_logging_config: StructLoggingConfig = field(default_factory=StructLoggingConfig)
    """Structlog Logging configuration for Litestar.  See ``litestar.logging.config.StructLoggingConfig``` for details."""
    middleware_logging_config: LoggingMiddlewareConfig = field(default_factory=LoggingMiddlewareConfig)
    """Middleware logging config."""
    enable_middleware_logging: bool = True
    """Enable request logging."""


class StructlogPlugin(InitPluginProtocol):
    """Structlog Plugin."""

    __slots__ = ("_config",)

    def __init__(self, config: StructlogConfig | None = None) -> None:
        if config is None:
            config = StructlogConfig()
        self._config = config
        super().__init__()

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        """Structlog Plugin

        Args:
            app_config: The :class:`AppConfig <litestar.config.app.AppConfig>` instance.

        Returns:
            The app config object.
        """
        if app_config.logging_config is not None and isinstance(app_config.logging_config, StructLoggingConfig):
            console.print(
                "[red dim]* Found pre-configured `StructLoggingConfig` on the `app` instance.  Skipping configuration.[/]",
            )
        else:
            app_config.logging_config = self._config.structlog_logging_config
        app_config.logging_config.configure()
        if self._config.structlog_logging_config.standard_lib_logging_config is not None:  # pragma: no cover
            self._config.structlog_logging_config.standard_lib_logging_config.configure()  # pragma: no cover
        if self._config.enable_middleware_logging:
            app_config.middleware.append(self._config.middleware_logging_config.middleware)
        return app_config  # pragma: no cover



from typing import Dict, List, Optional

import attrs

from litestar import post
from litestar.testing import create_test_client
from tests.models import DataclassPet


def test_spec_generation() -> None:
    @attrs.define
    class Person:
        first_name: str
        last_name: str
        id: str
        optional: Optional[str]
        complex: Dict[str, List[Dict[str, str]]]
        pets: Optional[List[DataclassPet]]

    @post("/")
    def handler(data: Person) -> Person:
        return data

    with create_test_client(handler) as client:
        schema = client.app.openapi_schema
        assert schema
        assert schema.to_schema()["components"]["schemas"]["test_spec_generation.Person"] == {
            "properties": {
                "first_name": {"type": "string"},
                "last_name": {"type": "string"},
                "id": {"type": "string"},
                "optional": {"oneOf": [{"type": "null"}, {"type": "string"}]},
                "complex": {
                    "type": "object",
                    "additionalProperties": {
                        "type": "array",
                        "items": {"type": "object", "additionalProperties": {"type": "string"}},
                    },
                },
                "pets": {
                    "oneOf": [
                        {"type": "null"},
                        {
                            "items": {"$ref": "#/components/schemas/DataclassPet"},
                            "type": "array",
                        },
                    ]
                },
            },
            "type": "object",
            "required": ["complex", "first_name", "id", "last_name"],
            "title": "Person",
        }



from __future__ import annotations

import dataclasses
from dataclasses import MISSING, fields
from typing import TYPE_CHECKING

from litestar.plugins import OpenAPISchemaPlugin
from litestar.types import Empty
from litestar.typing import FieldDefinition
from litestar.utils.predicates import is_optional_union

if TYPE_CHECKING:
    from litestar._openapi.schema_generation import SchemaCreator
    from litestar.openapi.spec import Schema


class DataclassSchemaPlugin(OpenAPISchemaPlugin):
    def is_plugin_supported_field(self, field_definition: FieldDefinition) -> bool:
        return field_definition.is_dataclass_type

    def to_openapi_schema(self, field_definition: FieldDefinition, schema_creator: SchemaCreator) -> Schema:
        type_hints = field_definition.get_type_hints(include_extras=True, resolve_generics=True)
        dataclass_fields = fields(field_definition.type_)
        return schema_creator.create_component_schema(
            field_definition,
            required=sorted(
                field.name
                for field in dataclass_fields
                if (
                    field.default is MISSING
                    and field.default_factory is MISSING
                    and not is_optional_union(type_hints[field.name])
                )
            ),
            property_fields={
                field.name: FieldDefinition.from_kwarg(
                    annotation=type_hints[field.name],
                    name=field.name,
                    default=field.default if field.default is not dataclasses.MISSING else Empty,
                )
                for field in dataclass_fields
            },
        )



from docs.examples.dependency_injection import dependency_with_dependency_fn_and_default

from litestar.status_codes import HTTP_200_OK
from litestar.testing import TestClient


def test_optional_dependency_not_in_openapi_schema() -> None:
    with TestClient(app=dependency_with_dependency_fn_and_default.app) as client:
        r = client.get("/schema/openapi.json")
    assert r.status_code == HTTP_200_OK
    assert r.json()["paths"]["/"]["get"].get("parameters") is None



from __future__ import annotations

from dataclasses import dataclass
from pathlib import PurePath  # noqa: TCH003
from typing import TYPE_CHECKING, Any, Sequence

from litestar.exceptions import ImproperlyConfiguredException
from litestar.file_system import BaseLocalFileSystem
from litestar.handlers import asgi, get, head
from litestar.response.file import ASGIFileResponse  # noqa: TCH001
from litestar.router import Router
from litestar.static_files.base import StaticFiles
from litestar.types import Empty
from litestar.utils import normalize_path, warn_deprecation

__all__ = ("StaticFilesConfig",)

if TYPE_CHECKING:
    from litestar.datastructures import CacheControlHeader
    from litestar.handlers.asgi_handlers import ASGIRouteHandler
    from litestar.openapi.spec import SecurityRequirement
    from litestar.types import (
        AfterRequestHookHandler,
        AfterResponseHookHandler,
        BeforeRequestHookHandler,
        EmptyType,
        ExceptionHandlersMap,
        Guard,
        Middleware,
        PathType,
    )


@dataclass
class StaticFilesConfig:
    """Configuration for static file service.

    To enable static files, pass an instance of this class to the :class:`Litestar <litestar.app.Litestar>` constructor using
    the 'static_files_config' key.
    """

    path: str
    """Path to serve static files from.

    Note that the path cannot contain path parameters.
    """
    directories: list[PathType]
    """A list of directories to serve files from."""
    html_mode: bool = False
    """Flag dictating whether serving html.

    If true, the default file will be 'index.html'.
    """
    name: str | None = None
    """An optional string identifying the static files handler."""
    file_system: Any = BaseLocalFileSystem()  # noqa: RUF009
    """The file_system spec to use for serving files.

    Notes:
        - A file_system is a class that adheres to the
            :class:`FileSystemProtocol <litestar.types.FileSystemProtocol>`.
        - You can use any of the file systems exported from the
            [fsspec](https://filesystem-spec.readthedocs.io/en/latest/) library for this purpose.
    """
    opt: dict[str, Any] | None = None
    """A string key dictionary of arbitrary values that will be added to the static files handler."""
    guards: list[Guard] | None = None
    """A list of :class:`Guard <litestar.types.Guard>` callables."""
    exception_handlers: ExceptionHandlersMap | None = None
    """A dictionary that maps handler functions to status codes and/or exception types."""
    send_as_attachment: bool = False
    """Whether to send the file as an attachment."""

    def __post_init__(self) -> None:
        _validate_config(path=self.path, directories=self.directories, file_system=self.file_system)
        self.path = normalize_path(self.path)
        warn_deprecation(
            "2.6.0",
            kind="class",
            deprecated_name="StaticFilesConfig",
            removal_in="3.0",
            alternative="create_static_files_router",
            info='Replace static_files_config=[StaticFilesConfig(path="/static", directories=["assets"])] with '
            'route_handlers=[..., create_static_files_router(path="/static", directories=["assets"])]',
        )

    def to_static_files_app(self) -> ASGIRouteHandler:
        """Return an ASGI app serving static files based on the config.

        Returns:
            :class:`StaticFiles <litestar.static_files.StaticFiles>`
        """
        static_files = StaticFiles(
            is_html_mode=self.html_mode,
            directories=self.directories,
            file_system=self.file_system,
            send_as_attachment=self.send_as_attachment,
        )
        return asgi(
            path=self.path,
            name=self.name,
            is_static=True,
            opt=self.opt,
            guards=self.guards,
            exception_handlers=self.exception_handlers,
        )(static_files)


def create_static_files_router(
    path: str,
    directories: list[PathType],
    file_system: Any = None,
    send_as_attachment: bool = False,
    html_mode: bool = False,
    name: str = "static",
    after_request: AfterRequestHookHandler | None = None,
    after_response: AfterResponseHookHandler | None = None,
    before_request: BeforeRequestHookHandler | None = None,
    cache_control: CacheControlHeader | None = None,
    exception_handlers: ExceptionHandlersMap | None = None,
    guards: list[Guard] | None = None,
    include_in_schema: bool | EmptyType = Empty,
    middleware: Sequence[Middleware] | None = None,
    opt: dict[str, Any] | None = None,
    security: Sequence[SecurityRequirement] | None = None,
    tags: Sequence[str] | None = None,
    router_class: type[Router] = Router,
    resolve_symlinks: bool = True,
) -> Router:
    """Create a router with handlers to serve static files.

    Args:
        path: Path to serve static files under
        directories: Directories to serve static files from
        file_system: A *file system* implementing
            :class:`~litestar.types.FileSystemProtocol`.
            `fsspec <https://filesystem-spec.readthedocs.io/en/latest/>`_ can be passed
            here as well
        send_as_attachment: Whether to send the file as an attachment
        html_mode: When in HTML:
            - Serve an ``index.html`` file from ``/``
            - Serve ``404.html`` when a file could not be found
        name: Name to pass to the generated handlers
        after_request: ``after_request`` handlers passed to the router
        after_response: ``after_response`` handlers passed to the router
        before_request: ``before_request`` handlers passed to the router
        cache_control: ``cache_control`` passed to the router
        exception_handlers: Exception handlers passed to the router
        guards: Guards  passed to the router
        include_in_schema: Include the routes / router in the OpenAPI schema
        middleware: Middlewares passed to the router
        opt: Opts passed to the router
        security: Security options passed to the router
        tags: ``tags`` passed to the router
        router_class: The class used to construct a router from
        resolve_symlinks: Resolve symlinks of ``directories``
    """

    if file_system is None:
        file_system = BaseLocalFileSystem()

    _validate_config(path=path, directories=directories, file_system=file_system)
    path = normalize_path(path)

    headers = None
    if cache_control:
        headers = {cache_control.HEADER_NAME: cache_control.to_header()}

    static_files = StaticFiles(
        is_html_mode=html_mode,
        directories=directories,
        file_system=file_system,
        send_as_attachment=send_as_attachment,
        resolve_symlinks=resolve_symlinks,
        headers=headers,
    )

    @get("{file_path:path}", name=name)
    async def get_handler(file_path: PurePath) -> ASGIFileResponse:
        return await static_files.handle(path=file_path.as_posix(), is_head_response=False)

    @head("/{file_path:path}", name=f"{name}/head")
    async def head_handler(file_path: PurePath) -> ASGIFileResponse:
        return await static_files.handle(path=file_path.as_posix(), is_head_response=True)

    handlers = [get_handler, head_handler]

    if html_mode:

        @get("/", name=f"{name}/index")
        async def index_handler() -> ASGIFileResponse:
            return await static_files.handle(path="/", is_head_response=False)

        handlers.append(index_handler)

    return router_class(
        after_request=after_request,
        after_response=after_response,
        before_request=before_request,
        cache_control=cache_control,
        exception_handlers=exception_handlers,
        guards=guards,
        include_in_schema=include_in_schema,
        middleware=middleware,
        opt=opt,
        path=path,
        route_handlers=handlers,
        security=security,
        tags=tags,
    )


def _validate_config(path: str, directories: list[PathType], file_system: Any) -> None:
    if not path:
        raise ImproperlyConfiguredException("path must be a non-zero length string,")

    if not directories or not any(bool(d) for d in directories):
        raise ImproperlyConfiguredException("directories must include at least one path.")

    if "{" in path:
        raise ImproperlyConfiguredException("path parameters are not supported for static files")

    if not (callable(getattr(file_system, "info", None)) and callable(getattr(file_system, "open", None))):
        raise ImproperlyConfiguredException("file_system must adhere to the FileSystemProtocol type")



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import Any, ForwardRef, Generic, List, Optional, Tuple, TypeVar, Union

import annotated_types
import msgspec
import pytest
from typing_extensions import Annotated, NotRequired, Required, TypedDict, get_type_hints

from litestar.exceptions import LitestarWarning
from litestar.params import DependencyKwarg, KwargDefinition, Parameter, ParameterKwarg
from litestar.typing import FieldDefinition, _unpack_predicate

from .test_utils.test_signature import T, _check_field_definition, field_definition_int, test_type_hints


@pytest.mark.parametrize(
    ("annotation", "expected"),
    [
        (
            int,
            {
                "raw": int,
                "annotation": int,
                "origin": None,
                "args": (),
                "metadata": (),
                "safe_generic_origin": None,
                "inner_types": (),
            },
        ),
        (
            List[int],
            {
                "raw": List[int],
                "annotation": List[int],
                "origin": list,
                "args": (int,),
                "metadata": (),
                "safe_generic_origin": List,
                "inner_types": (field_definition_int,),
            },
        ),
        (
            Annotated[int, "foo"],
            {
                "raw": Annotated[int, "foo"],
                "annotation": int,
                "origin": None,
                "args": (),
                "metadata": ("foo",),
                "safe_generic_origin": None,
                "inner_types": (),
            },
        ),
        (
            Annotated[List[int], "foo"],
            {
                "raw": Annotated[List[int], "foo"],
                "annotation": List[int],
                "origin": list,
                "args": (int,),
                "metadata": ("foo",),
                "safe_generic_origin": List,
                "inner_types": (field_definition_int,),
            },
        ),
        (
            test_type_hints["req_int"],
            {
                "raw": test_type_hints["req_int"],
                "annotation": int,
                "origin": None,
                "args": (),
                "metadata": (),
                "safe_generic_origin": None,
                "inner_types": (),
            },
        ),
        (
            test_type_hints["req_list_int"],
            {
                "raw": test_type_hints["req_list_int"],
                "annotation": List[int],
                "origin": list,
                "args": (int,),
                "metadata": (),
                "safe_generic_origin": List,
                "inner_types": (field_definition_int,),
            },
        ),
        (
            test_type_hints["not_req_int"],
            {
                "raw": test_type_hints["not_req_int"],
                "annotation": int,
                "origin": None,
                "args": (),
                "metadata": (),
                "safe_generic_origin": None,
                "inner_types": (),
            },
        ),
        (
            test_type_hints["not_req_list_int"],
            {
                "raw": test_type_hints["not_req_list_int"],
                "annotation": List[int],
                "origin": list,
                "args": (int,),
                "metadata": (),
                "safe_generic_origin": List,
                "inner_types": (field_definition_int,),
            },
        ),
        (
            test_type_hints["ann_req_int"],
            {
                "raw": test_type_hints["ann_req_int"],
                "annotation": int,
                "origin": None,
                "args": (),
                "metadata": ("foo",),
                "safe_generic_origin": None,
                "inner_types": (),
            },
        ),
        (
            test_type_hints["ann_req_list_int"],
            {
                "raw": test_type_hints["ann_req_list_int"],
                "annotation": List[int],
                "origin": list,
                "args": (int,),
                "metadata": ("foo",),
                "safe_generic_origin": List,
                "inner_types": (field_definition_int,),
            },
        ),
    ],
)
def test_field_definition_from_annotation(annotation: Any, expected: dict[str, Any]) -> None:
    """Test FieldDefinition.from_annotation."""
    _check_field_definition(FieldDefinition.from_annotation(annotation), expected)


def test_field_definition_kwarg_definition_from_extras() -> None:
    kwarg_definition = KwargDefinition()
    assert (
        FieldDefinition.from_annotation(int, extra={"kwarg_definition": kwarg_definition}).kwarg_definition
        is kwarg_definition
    )


@pytest.mark.parametrize("kwarg_definition", [KwargDefinition(), DependencyKwarg()])
def test_field_definition_kwarg_definition_from_kwargs(kwarg_definition: KwargDefinition | DependencyKwarg) -> None:
    assert FieldDefinition.from_annotation(int, kwarg_definition=kwarg_definition).kwarg_definition is kwarg_definition


def test_field_definition_with_annotated_kwarg_definition() -> None:
    kwarg_definition = KwargDefinition()
    fd = FieldDefinition.from_annotation(Annotated[str, kwarg_definition])
    assert fd.kwarg_definition is kwarg_definition


def test_field_definition_from_union_annotation() -> None:
    """Test FieldDefinition.from_annotation for Union."""
    annotation = Union[int, List[int]]
    expected = {
        "raw": annotation,
        "annotation": annotation,
        "origin": Union,
        "args": (int, List[int]),
        "metadata": (),
        "safe_generic_origin": Union,
        "inner_types": (FieldDefinition.from_annotation(int), FieldDefinition.from_annotation(List[int])),
    }
    _check_field_definition(FieldDefinition.from_annotation(annotation), expected)


@pytest.mark.parametrize("value", ["int", ForwardRef("int")])
def test_field_definition_is_forward_ref_predicate(value: Any) -> None:
    """Test FieldDefinition with ForwardRef."""
    field_definition = FieldDefinition.from_annotation(value)
    assert field_definition.is_forward_ref is True
    assert field_definition.annotation == value
    assert field_definition.origin is None
    assert field_definition.args == ()
    assert field_definition.metadata == ()
    assert field_definition.is_annotated is False
    assert field_definition.is_required is True
    assert field_definition.safe_generic_origin is None
    assert field_definition.inner_types == ()


def test_field_definition_is_type_var_predicate() -> None:
    """Test FieldDefinition.is_type_var."""
    assert FieldDefinition.from_annotation(int).is_type_var is False
    assert FieldDefinition.from_annotation(T).is_type_var is True
    assert FieldDefinition.from_annotation(Union[int, T]).is_type_var is False  # pyright: ignore


def test_field_definition_is_union_predicate() -> None:
    """Test FieldDefinition.is_union."""
    assert FieldDefinition.from_annotation(int).is_union is False
    assert FieldDefinition.from_annotation(Optional[int]).is_union is True
    assert FieldDefinition.from_annotation(Union[int, None]).is_union is True
    assert FieldDefinition.from_annotation(Union[int, str]).is_union is True


def test_field_definition_is_optional_predicate() -> None:
    """Test FieldDefinition.is_optional."""
    assert FieldDefinition.from_annotation(int).is_optional is False
    assert FieldDefinition.from_annotation(Optional[int]).is_optional is True
    assert FieldDefinition.from_annotation(Union[int, None]).is_optional is True
    assert FieldDefinition.from_annotation(Union[int, None, str]).is_optional is True
    assert FieldDefinition.from_annotation(Union[int, str]).is_optional is False


def test_field_definition_is_dataclass_predicate() -> None:
    """Test FieldDefinition.is_dataclass."""

    class NormalClass: ...

    @dataclass
    class NormalDataclass: ...

    @dataclass
    class GenericDataclass(Generic[T]): ...

    assert FieldDefinition.from_annotation(NormalDataclass).is_dataclass_type is True
    assert FieldDefinition.from_annotation(GenericDataclass).is_dataclass_type is True
    assert FieldDefinition.from_annotation(GenericDataclass[int]).is_dataclass_type is True
    assert FieldDefinition.from_annotation(GenericDataclass[T]).is_dataclass_type is True  # type: ignore[valid-type]
    assert FieldDefinition.from_annotation(NormalClass).is_dataclass_type is False


def test_field_definition_is_typeddict_predicate() -> None:
    """Test FieldDefinition.is_typeddict."""

    class NormalClass: ...

    class TypedDictClass(TypedDict): ...

    assert FieldDefinition.from_annotation(NormalClass).is_typeddict_type is False
    assert FieldDefinition.from_annotation(TypedDictClass).is_typeddict_type is True

    if sys.version_info >= (3, 11):

        class GenericTypedDictClass(TypedDict, Generic[T]): ...

        assert FieldDefinition.from_annotation(GenericTypedDictClass).is_typeddict_type is True
        assert FieldDefinition.from_annotation(GenericTypedDictClass[int]).is_typeddict_type is True
        assert FieldDefinition.from_annotation(GenericTypedDictClass[T]).is_typeddict_type is True


def test_field_definition_is_subclass_of() -> None:
    """Test FieldDefinition.is_type_of."""
    assert FieldDefinition.from_annotation(bool).is_subclass_of(int) is True
    assert FieldDefinition.from_annotation(bool).is_subclass_of(str) is False
    assert FieldDefinition.from_annotation(Union[int, str]).is_subclass_of(int) is False
    assert FieldDefinition.from_annotation(List[int]).is_subclass_of(list) is True
    assert FieldDefinition.from_annotation(List[int]).is_subclass_of(int) is False
    assert FieldDefinition.from_annotation(Optional[int]).is_subclass_of(int) is False
    assert FieldDefinition.from_annotation(Union[bool, int]).is_subclass_of(int) is True


def test_field_definition_has_inner_subclass_of() -> None:
    """Test FieldDefinition.has_type_of."""
    assert FieldDefinition.from_annotation(List[int]).has_inner_subclass_of(int) is True
    assert FieldDefinition.from_annotation(List[int]).has_inner_subclass_of(str) is False
    assert FieldDefinition.from_annotation(List[Union[int, str]]).has_inner_subclass_of(int) is False


def test_field_definition_equality() -> None:
    assert FieldDefinition.from_annotation(int) == FieldDefinition.from_annotation(int)
    assert FieldDefinition.from_annotation(int) == FieldDefinition.from_annotation(Annotated[int, "meta"])
    assert FieldDefinition.from_annotation(int) != int
    assert FieldDefinition.from_annotation(List[int]) == FieldDefinition.from_annotation(List[int])
    assert FieldDefinition.from_annotation(List[int]) != FieldDefinition.from_annotation(List[str])
    assert FieldDefinition.from_annotation(List[str]) != FieldDefinition.from_annotation(Tuple[str])
    assert FieldDefinition.from_annotation(Optional[str]) == FieldDefinition.from_annotation(Union[str, None])


def test_field_definition_hash() -> None:
    assert hash(FieldDefinition.from_annotation(int)) == hash(FieldDefinition.from_annotation(int))
    assert hash(FieldDefinition.from_annotation(Annotated[int, False])) == hash(
        FieldDefinition.from_annotation(Annotated[int, False])
    )
    assert hash(FieldDefinition.from_annotation(Annotated[int, False])) != hash(
        FieldDefinition.from_annotation(Annotated[int, True])
    )
    assert hash(FieldDefinition.from_annotation(Union[str, int])) != hash(
        FieldDefinition.from_annotation(Union[int, str])
    )


def test_is_required() -> None:
    class Foo(TypedDict):
        required: Required[str]
        not_required: NotRequired[str]

    class Bar(msgspec.Struct):
        unset: Union[str, msgspec.UnsetType] = msgspec.UNSET  # noqa: UP007
        with_default: str = ""
        with_none_default: Union[str, None] = None  # noqa: UP007

    assert FieldDefinition.from_annotation(get_type_hints(Foo, include_extras=True)["required"]).is_required is True
    assert (
        FieldDefinition.from_annotation(get_type_hints(Foo, include_extras=True)["not_required"]).is_required is False
    )
    assert FieldDefinition.from_annotation(get_type_hints(Bar, include_extras=True)["unset"]).is_required is False

    assert (
        FieldDefinition.from_kwarg(
            name="foo", kwarg_definition=ParameterKwarg(required=False), annotation=str
        ).is_required
        is False
    )
    assert (
        FieldDefinition.from_kwarg(
            name="foo", kwarg_definition=ParameterKwarg(required=True), annotation=str
        ).is_required
        is True
    )
    assert (
        FieldDefinition.from_kwarg(
            name="foo", kwarg_definition=ParameterKwarg(required=None, default=""), annotation=str
        ).is_required
        is False
    )
    assert (
        FieldDefinition.from_kwarg(
            name="foo", kwarg_definition=ParameterKwarg(required=None), annotation=str
        ).is_required
        is True
    )

    assert FieldDefinition.from_annotation(Optional[str]).is_required is False
    assert FieldDefinition.from_annotation(str).is_required is True

    assert FieldDefinition.from_annotation(Any).is_required is False

    assert FieldDefinition.from_annotation(get_type_hints(Bar)["with_default"]).is_required is True
    assert FieldDefinition.from_annotation(get_type_hints(Bar)["with_none_default"]).is_required is False


def test_field_definition_bound_type() -> None:
    class Foo:
        pass

    class Bar:
        pass

    bound = TypeVar("bound", bound=Foo)
    multiple_bounds = TypeVar("multiple_bounds", bound=Union[Foo, Bar])

    assert FieldDefinition.from_annotation(str).bound_types is None
    assert FieldDefinition.from_annotation(T).bound_types is None

    bound_types = FieldDefinition.from_annotation(bound).bound_types

    assert bound_types
    assert len(bound_types) == 1
    assert isinstance(bound_types[0], FieldDefinition)
    assert bound_types[0].raw is Foo

    bound_types_union = FieldDefinition.from_annotation(multiple_bounds).bound_types
    assert bound_types_union
    assert len(bound_types_union) == 2
    assert bound_types_union[0].raw is Foo
    assert bound_types_union[1].raw is Bar


def test_nested_generic_types() -> None:
    V = TypeVar("V")

    class Foo(Generic[T]):
        pass

    class Bar(Generic[T, V]):
        pass

    class Baz(Generic[T], Bar[T, str]):
        pass

    fd_simple = FieldDefinition.from_annotation(Foo)
    assert fd_simple.generic_types
    assert len(fd_simple.generic_types) == 1
    assert fd_simple.generic_types[0].raw == T

    fd_union = FieldDefinition.from_annotation(Bar)
    assert fd_union.generic_types
    assert len(fd_union.generic_types) == 2
    assert fd_union.generic_types[0].raw == T
    assert fd_union.generic_types[1].raw == V

    fd_nested = FieldDefinition.from_annotation(Baz)
    assert fd_nested.generic_types
    assert len(fd_nested.generic_types) == 3
    assert fd_nested.generic_types[0].raw == T
    assert fd_nested.generic_types[1].raw == T
    assert fd_nested.generic_types[2].raw == str


@dataclass
class GenericDataclass(Generic[T]):
    foo: T


@dataclass
class NormalDataclass:
    foo: int


@pytest.mark.parametrize(
    ("annotation", "expected_type_hints"),
    ((GenericDataclass[str], {"foo": str}), (GenericDataclass, {"foo": T}), (NormalDataclass, {"foo": int})),
)
def test_field_definition_get_type_hints(annotation: Any, expected_type_hints: dict[str, Any]) -> None:
    assert (
        FieldDefinition.from_annotation(annotation).get_type_hints(include_extras=True, resolve_generics=True)
        == expected_type_hints
    )


@pytest.mark.parametrize(
    ("annotation", "expected_type_hints"),
    ((GenericDataclass[str], {"foo": T}), (GenericDataclass, {"foo": T}), (NormalDataclass, {"foo": int})),
)
def test_field_definition_get_type_hints_dont_resolve_generics(
    annotation: Any, expected_type_hints: dict[str, Any]
) -> None:
    assert (
        FieldDefinition.from_annotation(annotation).get_type_hints(include_extras=True, resolve_generics=False)
        == expected_type_hints
    )


@pytest.mark.parametrize(
    "predicate, expected_meta",
    [
        (annotated_types.LowerCase.__metadata__[0], {"lower_case": True}),  # pyright: ignore
        (annotated_types.UpperCase.__metadata__[0], {"upper_case": True}),  # pyright: ignore
        (annotated_types.IsAscii.__metadata__[0], {"pattern": "[[:ascii:]]"}),  # pyright: ignore
        (annotated_types.IsDigits.__metadata__[0], {"pattern": "[[:digit:]]"}),  # pyright: ignore
        (object(), {}),
    ],
)
def test_unpack_predicate(predicate: Any, expected_meta: dict[str, Any]) -> None:
    assert _unpack_predicate(predicate) == expected_meta


def test_warn_ambiguous_default_values() -> None:
    with pytest.warns(LitestarWarning, match="Ambiguous default values"):
        FieldDefinition.from_annotation(Annotated[int, Parameter(default=1)], default=2)


def test_warn_defaults_inside_parameter_definition() -> None:
    with pytest.warns(DeprecationWarning, match="Deprecated default value specification"):
        FieldDefinition.from_annotation(Annotated[int, Parameter(default=1)], default=1)



from litestar.middleware.compression.facade import CompressionFacade
from litestar.middleware.compression.middleware import CompressionMiddleware

__all__ = ("CompressionMiddleware", "CompressionFacade")



from __future__ import annotations

from dataclasses import dataclass, fields, is_dataclass
from typing import TYPE_CHECKING, Any, Hashable, Mapping, Sequence

from litestar.openapi.spec.base import BaseSchemaObject
from litestar.utils.predicates import is_non_string_sequence

if TYPE_CHECKING:
    from litestar.openapi.spec.discriminator import Discriminator
    from litestar.openapi.spec.enums import OpenAPIFormat, OpenAPIType
    from litestar.openapi.spec.external_documentation import ExternalDocumentation
    from litestar.openapi.spec.reference import Reference
    from litestar.openapi.spec.xml import XML
    from litestar.types import DataclassProtocol

__all__ = ("Schema", "SchemaDataContainer")


def _recursive_hash(value: Hashable | Sequence | Mapping | DataclassProtocol | type[DataclassProtocol]) -> int:
    if isinstance(value, Mapping):
        hash_value = 0
        for k, v in value.items():
            if k != "examples":
                hash_value += hash(k)
                hash_value += _recursive_hash(v)
        return hash_value
    if is_dataclass(value):
        hash_value = hash(type(value).__name__)
        for field in fields(value):
            if field.name != "examples":
                hash_value += hash(field.name)
                hash_value += _recursive_hash(getattr(value, field.name, None))
        return hash_value
    if is_non_string_sequence(value):
        return sum(_recursive_hash(v) for v in value)
    return hash(value) if isinstance(value, Hashable) else 0


@dataclass
class Schema(BaseSchemaObject):
    """The Schema Object allows the definition of input and output data types. These types can be objects, but also
    primitives and arrays. This object is a superset of the
    `JSON Schema Specification Draft 2020-12 <https://tools.ietf.org/html/draft-bhutton-json-schema-00>`_.

    For more information about the properties, see
    `JSON Schema Core <https://tools.ietf.org/html/draft-wright-json-schema-00>`_ and
    `JSON Schema Validation <https://tools.ietf.org/html/draft-wright-json-schema-validation-00>`_.

    Unless stated otherwise, the property definitions follow those of JSON Schema and do not add any additional
    semantics. Where JSON Schema indicates that behavior is defined by the application (e.g. for annotations), OAS also
    defers the definition of semantics to the application consuming the OpenAPI document.

    The following properties are taken directly from the
    `JSON Schema Core <https://tools.ietf.org/html/draft-wright-json-schema-00>`_ and follow the same specifications.
    """

    all_of: Sequence[Reference | Schema] | None = None
    """This keyword's value MUST be a non-empty array.  Each item of the array MUST be a valid JSON Schema.

    An instance validates successfully against this keyword if it validates successfully against all schemas defined by
    this keyword's value.
    """

    any_of: Sequence[Reference | Schema] | None = None
    """This keyword's value MUST be a non-empty array. Each item of the array MUST be a valid JSON Schema.

    An instance validates successfully against this keyword if it validates successfully against at least one schema
    defined by this keyword's value.  Note that when annotations are being collected, all subschemas MUST be examined so
    that annotations are collected from each subschema that validates successfully.
    """

    one_of: Sequence[Reference | Schema] | None = None
    """This keyword's value MUST be a non-empty array.  Each item of the array MUST be a valid JSON Schema.

    An instance validates successfully against this keyword if it validates successfully against exactly one schema
    defined by this keyword's value.
    """

    schema_not: Reference | Schema | None = None
    """This keyword's value MUST be a valid JSON Schema.

    An instance is valid against this keyword if it fails to validate successfully against the schema defined by this
    keyword.
    """

    schema_if: Reference | Schema | None = None
    """This keyword's value MUST be a valid JSON Schema.

    This validation outcome of this keyword's subschema has no direct effect on the overall validation result. Rather,
    it controls which of the "then" or "else" keywords are evaluated.

    Instances that successfully validate against this keyword's subschema MUST also be valid against the subschema
    value of the "then" keyword, if present.

    Instances that fail to validate against this keyword's subschema MUST also be valid against the subschema value of
    the "else" keyword, if present.

    If annotations (Section 7.7) are being collected, they are collected rom this keyword's subschema in the usual way,
    including when the keyword is present without either "then" or "else".
    """

    then: Reference | Schema | None = None
    """This keyword's value MUST be a valid JSON Schema.

    When "if" is present, and the instance successfully validates against its subschema, then validation succeeds
    against this keyword if the instance also successfully validates against this keyword's subschema.

    This keyword has no effect when "if" is absent, or when the instance fails to validate against its subschema.
    Implementations MUST NOT evaluate the instance against this keyword, for either validation or annotation collection
    purposes, in such cases.
    """

    schema_else: Reference | Schema | None = None
    """This keyword's value MUST be a valid JSON Schema.

    When "if" is present, and the instance fails to validate against its subschema, then validation succeeds against
    this keyword if the instance successfully validates against this keyword's subschema.

    This keyword has no effect when "if" is absent, or when the instance successfully validates against its subschema.
    Implementations MUST NOT evaluate the instance against this keyword, for either validation or annotation collection
    purposes, in such cases.
    """

    dependent_schemas: dict[str, Reference | Schema] | None = None
    """This keyword specifies subschemas that are evaluated if the instance is
    an object and contains a certain property.

    This keyword's value MUST be an object.  Each value in the object MUST be a valid JSON Schema.

    If the object key is a property in the instance, the entire instance must validate against the subschema. Its use is
    dependent on the presence of the property.

    Omitting this keyword has the same behavior as an empty object.
    """

    prefix_items: Sequence[Reference | Schema] | None = None
    """The value of "prefixItems" MUST be a non-empty array of valid JSON Schemas.

    Validation succeeds if each element of the instance validates against the schema at the same position, if any.
    This keyword does not constrain the length of the array.  If the array is longer than this keyword's value, this
    keyword validates only the prefix of matching length.

    This keyword produces an annotation value which is the largest index to which this keyword applied a subschema.
    he value MAY be a boolean true if a subschema was applied to every index of the instance, such as is produced by the
    "items" keyword.  This annotation affects the behavior of "items" and "unevaluatedItems".

    Omitting this keyword has the same assertion behavior as an empty array.
    """

    items: Reference | Schema | None = None
    """The value of "items" MUST be a valid JSON Schema.

    This keyword applies its subschema to all instance elements at indexes greater than the length of the "prefixItems"
    array in the same schema object, as reported by the annotation result of that "prefixItems" keyword.  If no such
    annotation result exists, "items" applies its subschema to all instance array elements.  [[CREF11: Note that the
    behavior of "items" without "prefixItems" is identical to that of the schema form of "items" in prior drafts. When
    "prefixItems" is present, the behavior of "items" is identical to the former "additionalItems" keyword.  ]]

    If the "items" subschema is applied to any positions within the instance array, it produces an annotation result of
    boolean true, indicating that all remaining array elements have been evaluated against this keyword's subschema.

    Omitting this keyword has the same assertion behavior as an empty schema.

    Implementations MAY choose to implement or optimize this keyword in another way that produces the same effect, such
    as by directly checking for the presence and size of a "prefixItems" array. Implementations that do not support
    annotation collection MUST do so.
    """

    contains: Reference | Schema | None = None
    """The value of this keyword MUST be a valid JSON Schema.

    An array instance is valid against "contains" if at least one of its elements is valid against the given schema.
    The subschema MUST be applied to every array element even after the first match has been found, in order to collect
    annotations for use by other keywords. This is to ensure that all possible annotations are collected.

    Logically, the validation result of applying the value subschema to each item in the array MUST be ORed with
    "false", resulting in an overall validation result.

    This keyword produces an annotation value which is an array of the indexes to which this keyword validates
    successfully when applying its subschema, in ascending order.  The value MAY be a boolean "true" if the subschema
    validates successfully when applied to every index of the instance.  The annotation MUST be present if the instance
    array to which this keyword's schema applies is empty.
    """

    properties: dict[str, Reference | Schema] | None = None
    """The value of "properties" MUST be an object.  Each value of this object MUST be a valid JSON Schema.

    Validation succeeds if, for each name that appears in both the instance and as a name within this keyword's value,
    the child instance for that name successfully validates against the corresponding schema.

    The annotation result of this keyword is the set of instance property names matched by this keyword.

    Omitting this keyword has the same assertion behavior as an empty object.
    """

    pattern_properties: dict[str, Reference | Schema] | None = None
    """The value of "patternProperties" MUST be an object.  Each property name of this object SHOULD be a valid
    regular expression, according to the ECMA-262 regular expression dialect.  Each property value of this object
    MUST be a valid JSON Schema.

    Validation succeeds if, for each instance name that matches any regular expressions that appear as a property name
    in this keyword's value, the child instance for that name successfully validates against each schema that
    corresponds to a matching regular expression.

    The annotation result of this keyword is the set of instance property names matched by this keyword.

    Omitting this keyword has the same assertion behavior as an empty object.
    """

    additional_properties: Reference | Schema | bool | None = None
    """The value of "additionalProperties" MUST be a valid JSON Schema.

    The behavior of this keyword depends on the presence and annotation results of "properties" and "patternProperties"
    within the same schema object.  Validation with "additionalProperties" applies only to the child values of instance
    names that do not appear in the annotation results of either "properties" or "patternProperties".

    For all such properties, validation succeeds if the child instance validates against the "additionalProperties"
    schema.

    The annotation result of this keyword is the set of instance property names validated by this keyword's subschema.

    Omitting this keyword has the same assertion behavior as an empty schema.

    Implementations MAY choose to implement or optimize this keyword in another way that produces the same effect, such
    as by directly checking the names in "properties" and the patterns in "patternProperties" against the instance
    property set. Implementations that do not support annotation collection MUST do so.
    """

    property_names: Reference | Schema | None = None
    """The value of "propertyNames" MUST be a valid JSON Schema.

    If the instance is an object, this keyword validates if every property name in the instance validates against the
    provided schema. Note the property name that the schema is testing will always be a string.

    Omitting this keyword has the same behavior as an empty schema.
    """

    unevaluated_items: Reference | Schema | None = None
    """The value of "unevaluatedItems" MUST be a valid JSON Schema.

    The behavior of this keyword depends on the annotation results of adjacent keywords that apply to the instance
    location being validated.  Specifically, the annotations from "prefixItems" items", and "contains", which can come
    from those keywords when they are adjacent to the "unevaluatedItems" keyword.  Those three annotations, as well as
    "unevaluatedItems", can also result from any and all adjacent in-place applicator (Section 10.2) keywords.  This
    includes but is not limited to the in-place applicators defined in this document.

    If no relevant annotations are present, the "unevaluatedItems" subschema MUST be applied to all locations in the
    array. If a boolean true value is present from any of the relevant annotations, unevaluatedItems" MUST be ignored.
    Otherwise, the subschema MUST be applied to any index greater than the largest annotation value for "prefixItems",
    which does not appear in any annotation value for
    "contains".

    This means that "prefixItems", "items", "contains", and all in-place applicators MUST be evaluated before this
    keyword can be evaluated. Authors of extension keywords MUST NOT define an in-place applicator that would need to be
    evaluated after this keyword.

    If the "unevaluatedItems" subschema is applied to any positions within the instance array, it produces an annotation
    result of boolean true, analogous to the behavior of "items".

    Omitting this keyword has the same assertion behavior as an empty schema.
    """

    unevaluated_properties: Reference | Schema | None = None
    """The value of "unevaluatedProperties" MUST be a valid JSON Schema.

    The behavior of this keyword depends on the annotation results of adjacent keywords that apply to the instance
    location being validated.  Specifically, the annotations from "properties", "patternProperties", and
    "additionalProperties", which can come from those keywords when they are adjacent to the "unevaluatedProperties"
    keyword. Those three annotations, as well as "unevaluatedProperties", can also result from any and all adjacent
    in-place applicator (Section 10.2) keywords.  This includes but is not limited to the in-place applicators defined
    in this document.

    Validation with "unevaluatedProperties" applies only to the child values of instance names that do not appear in
    the "properties", "patternProperties", "additionalProperties", or "unevaluatedProperties" annotation results that
    apply to the instance location being validated.

    For all such properties, validation succeeds if the child instance validates against the "unevaluatedProperties"
    schema.

    This means that "properties", "patternProperties", "additionalProperties", and all in-place applicators MUST be
    evaluated before this keyword can be evaluated.  Authors of extension keywords MUST NOT define an in-place
    applicator that would need to be evaluated after this keyword.

    The annotation result of this keyword is the set of instance property names validated by this keyword's subschema.

    Omitting this keyword has the same assertion behavior as an empty schema.

    The following properties are taken directly from the
    `JSON Schema Validation <https://tools.ietf.org/html/draft-wright-json-schema-validation-00>`_ and follow the same
    specifications:
    """

    type: OpenAPIType | Sequence[OpenAPIType] | None = None
    """The value of this keyword MUST be either a string or an array.  If it is an array, elements of the array MUST be
    strings and MUST be unique.

    String values MUST be one of the six primitive types (``"null"``, ``"boolean"``, ``"object"``, ``"array"``,
    ``"number"``, and ``"string"``), or ``"integer"`` which matches any number with a zero fractional part.

    An instance validates if and only if the instance is in any of the sets listed for this keyword.
    """

    enum: Sequence[Any] | None = None
    """The value of this keyword MUST be an array.  This array SHOULD have at least one element.  Elements in the array
    SHOULD be unique.

    An instance validates successfully against this keyword if its value is equal to one of the elements in this
    keyword's array value.

    Elements in the array might be of any type, including null.
    """

    const: Any | None = None
    """The value of this keyword MAY be of any type, including null.

    Use of this keyword is functionally equivalent to an "enum" (Section 6.1.2) with a single value.

    An instance validates successfully against this keyword if its value is equal to the value of the keyword.
    """

    multiple_of: float | None = None
    """The value of "multipleOf" MUST be a number, strictly greater than 0.

    A numeric instance is only valid if division by this keyword's value results in an integer.
    """

    maximum: float | None = None
    """The value of "maximum" MUST be a number, representing an inclusive upper limit for a numeric instance.

    If the instance is a number, then this keyword validates only if the instance is less than or exactly equal to
    "maximum".
    """

    exclusive_maximum: float | None = None
    """The value of "exclusiveMaximum" MUST be a number, representing an exclusive upper limit for a numeric instance.

    If the instance is a number, then the instance is valid only if it has a value strictly less than (not equal to)
    "exclusiveMaximum".
    """

    minimum: float | None = None
    """The value of "minimum" MUST be a number, representing an inclusive lower limit for a numeric instance.

    If the instance is a number, then this keyword validates only if the instance is greater than or exactly equal to
    "minimum".
    """

    exclusive_minimum: float | None = None
    """The value of "exclusiveMinimum" MUST be a number, representing an exclusive lower limit for a numeric instance.

    If the instance is a number, then the instance is valid only if it has a value strictly greater than (not equal to)
    "exclusiveMinimum".
    """

    max_length: int | None = None
    """The value of this keyword MUST be a non-negative integer.

    A string instance is valid against this keyword if its length is less than, or equal to, the value of this keyword.

    The length of a string instance is defined as the number of its characters as defined by :rfc:`8259`.
    """

    min_length: int | None = None
    """The value of this keyword MUST be a non-negative integer.

    A string instance is valid against this keyword if its length is greater than, or equal to, the value of this
    keyword.

    The length of a string instance is defined as the number of its characters as defined by :rfc:`8259`.

    Omitting this keyword has the same behavior as a value of 0.
    """

    pattern: str | None = None
    """The value of this keyword MUST be a string.  This string SHOULD be a valid regular expression, according to the
    ECMA-262 regular expression dialect.

    A string instance is considered valid if the regular expression matches the instance successfully.  Recall: regular
    expressions are not implicitly anchored.
    """

    max_items: int | None = None
    """The value of this keyword MUST be a non-negative integer.

    An array instance is valid against "maxItems" if its size is less than, or equal to, the value of this keyword.
    """

    min_items: int | None = None
    """The value of this keyword MUST be a non-negative integer.

    An array instance is valid against "minItems" if its size is greater than, or equal to, the value of this keyword.

    Omitting this keyword has the same behavior as a value of 0.
    """

    unique_items: bool | None = None
    """The value of this keyword MUST be a boolean.

    If this keyword has boolean value false, the instance validates successfully.  If it has boolean value true, the
    instance validates successfully if all of its elements are unique.

    Omitting this keyword has the same behavior as a value of false.
    """

    max_contains: int | None = None
    """The value of this keyword MUST be a non-negative integer.

    If "contains" is not present within the same schema object, then this keyword has no effect.

    An instance array is valid against "maxContains" in two ways, depending on the form of the annotation result of an
    adjacent "contains" [json-schema] keyword.  The first way is if the annotation result is an array and the length of
    that array is less than or equal to the "maxContains" value.  The second way is if the annotation result is a
    boolean "true" and the instance array length is less than r equal to the "maxContains" value.
    """

    min_contains: int | None = None
    """The value of this keyword MUST be a non-negative integer.

    If "contains" is not present within the same schema object, then this keyword has no effect.

    An instance array is valid against "minContains" in two ways, depending on the form of the annotation result of an
    adjacent "contains" [json-schema] keyword.  The first way is if the annotation result is an array and the length of
    that array is greater than or equal to the "minContains" value.  The second way is if the annotation result is a
    boolean "true" and the instance array length is greater than or equal to the "minContains" value.

    A value of 0 is allowed, but is only useful for setting a range of occurrences from 0 to the value of "maxContains".
    A value of 0 with no "maxContains" causes "contains" to always pass validation.

    Omitting this keyword has the same behavior as a value of 1.
    """

    max_properties: int | None = None
    """The value of this keyword MUST be a non-negative integer.

    An object instance is valid against "maxProperties" if its number of properties is less than, or equal to, the value
    of this keyword.
    """

    min_properties: int | None = None
    """The value of this keyword MUST be a non-negative integer.

    An object instance is valid against "minProperties" if its number of properties is greater than, or equal to, the
    value of this keyword.

    Omitting this keyword has the same behavior as a value of 0.
    """

    required: Sequence[str] | None = None
    """The value of this keyword MUST be an array.  Elements of this array, if any, MUST be strings, and MUST be unique.

    An object instance is valid against this keyword if every item in the rray is the name of a property in the instance.

    Omitting this keyword has the same behavior as an empty array.
    """

    dependent_required: dict[str, Sequence[str]] | None = None
    """The value of this keyword MUST be an object.  Properties in this object, f any, MUST be arrays.  Elements in each
    array, if any, MUST be strings, and MUST be unique.

    This keyword specifies properties that are required if a specific other property is present.  Their requirement is
    dependent on the presence of the other property.

    Validation succeeds if, for each name that appears in both the instance and as a name within this keyword's value,
    every item in the corresponding array is also the name of a property in the instance.

    Omitting this keyword has the same behavior as an empty object.
    """

    format: OpenAPIFormat | None = None
    """From OpenAPI:

    See `Data Type Formats <https://spec.openapis.org/oas/v3.1.0#dataTypeFormat>`_ for further details. While relying on
    JSON Schema's defined formats, the OAS offers a few additional predefined formats.

    From JSON Schema:

    Structural validation alone may be insufficient to allow an application to correctly utilize certain values.
    The "format" annotation keyword is defined to allow schema authors to convey semantic information for a fixed subset
    of values which are accurately described by authoritative resources, be they RFCs or other external specifications.

    The value of this keyword is called a format attribute.  It MUST be a string.  A format attribute can generally only
    validate a given set of instance types.  If the type of the instance to validate is not in this set, validation for
    this format attribute and instance SHOULD succeed.  All format attributes defined in this section apply to strings,
    but a format attribute can be specified to apply to any instance types defined in the data model defined in the core
    JSON Schema. [json-schema] [[CREF1: Note that the "type" keyword in this specification defines an "integer" type
    which is not part of the data model. Therefore a format attribute can be limited to numbers, but not specifically to
    integers.  However, a numeric format can be used alongside the "type" keyword with a value of "integer", or could be
    explicitly defined to always pass if the number is not an integer, which produces essentially the same behavior as
    only applying to integers.  ]]
    """

    content_encoding: str | None = None
    """If the instance value is a string, this property defines that the string SHOULD be interpreted as binary data and
    decoded using the encoding named by this property.

    Possible values indicating base 16, 32, and 64 encodings with several variations are listed in :rfc:`4648`.
    Additionally, sections 6.7 and 6.8 of :rfc:`2045` provide encodings used in MIME.  As "base64" is defined in both
    RFCs, the definition from :rfc:`4648` SHOULD be assumed unless the string is specifically intended for use in a
    MIME context.  Note that all of these encodings result in strings consisting only of 7-bit ASCII characters.
    therefore, this keyword has no meaning for strings containing characters outside of that range.

    If this keyword is absent, but "contentMediaType" is present, this indicates that the encoding is the identity
    encoding, meaning that no transformation was needed in order to represent the content in a UTF-8 string.
    """

    content_media_type: str | None = None
    """If the instance is a string, this property indicates the media type of the contents of the string. If
    "contentEncoding" is present, this property describes the decoded string.

    The value of this property MUST be a string, which MUST be a media type, as defined by :rfc:`2046`
    """

    content_schema: Reference | Schema | None = None
    """If the instance is a string, and if "contentMediaType" is present, this property contains a schema which
    describes the structure of the string.

    This keyword MAY be used with any media type that can be mapped into JSON Schema's data model.

    The value of this property MUST be a valid JSON schema. It SHOULD be ignored if "contentMediaType" is not present.
    """

    title: str | None = None
    """The value of "title" MUST be a string.

    The title can be used to decorate a user interface with information about the data produced by this user interface.
    A title will preferably be short.
    """

    description: str | None = None
    """From OpenAPI:

    `CommonMark syntax <https://spec.commonmark.org/>`_ MAY be used for rich text representation.

    From JSON Schema:
    The value "description" MUST be a string.

    The description can be used to decorate a user interface with information about the data produced by this user
    interface. A description will provide explanation about the purpose of the instance described by this schema.
    """

    default: Any | None = None
    """There are no restrictions placed on the value of this keyword.  When multiple occurrences of this keyword are
    applicable to a single sub-instance, implementations SHOULD remove duplicates.

    This keyword can be used to supply a default JSON value associated with a particular schema.  It is RECOMMENDED that
    a default value be valid against the associated schema.
    """

    deprecated: bool | None = None
    """The value of this keyword MUST be a boolean.  When multiple occurrences of this keyword are applicable to a
    single sub-instance, applications SHOULD consider the instance location to be deprecated if any occurrence specifies
    a true value.

    If "deprecated" has a value of boolean true, it indicates that applications SHOULD refrain from usage of the
    declared property.  It  MAY mean the property is going to be removed in the future.

    A root schema containing "deprecated" with a value of true indicates that the entire resource being described MAY be
    removed in the future.

    The "deprecated" keyword applies to each instance location to which the schema object containing the keyword
    successfully applies. This  can result in scenarios where every array item or object property is deprecated even
    though the containing array or object is not.

    Omitting this keyword has the same behavior as a value of false.
    """

    read_only: bool | None = None
    """The value of "readOnly" MUST be a boolean.  When multiple occurrences of this keyword are applicable to a single
    sub-instance, the resulting behavior SHOULD be as for a true value if any occurrence specifies a true value, and
    SHOULD be as for a false value otherwise.

    If "readOnly" has a value of boolean true, it indicates that the value of the instance is managed exclusively by
    the owning authority, and attempts by an application to modify the value of this property are expected to be ignored
    or rejected by that owning authority.

    An instance document that is marked as "readOnly" for the entire document MAY be ignored if sent to the owning
    authority, or MAY result in an error, at the authority's discretion.

    For example, "readOnly" would be used to mark a database-generated serial number as read-only, while "writeOnly"
    would be used to mark a password input field.

    This keyword can be used to assist in user interface instance generation.  In particular, an application MAY choose
    to use a widget that hides input values as they are typed for write-only fields.

    Omitting these keywords has the same behavior as values of false.
    """

    write_only: bool | None = None
    """The value of "writeOnly" MUST be a boolean.  When multiple occurrences of this keyword are applicable to a
    single sub-instance, the resulting behavior SHOULD be as for a true value if any occurrence specifies a true value,
    and SHOULD be as for a false value otherwise.

    If "writeOnly" has a value of boolean true, it indicates that the value is never present when the instance is
    retrieved from the owning authority.  It can be present when sent to the owning authority to update or create the
    document (or the resource it represents), but it will not be included in any updated or newly created version of the
    instance.

    An instance document that is marked as "writeOnly" for the entire document MAY be returned as a blank document of
    some sort, or MAY produce an error upon retrieval, or have the retrieval request ignored, at the authority's
    discretion.

    For example, "readOnly" would be used to mark a database-generated serial number as read-only, while "writeOnly"
    would be used to mark a  password input field.

    This keyword can be used to assist in user interface instance generation. In particular, an application MAY choose
    to use a widget that hides input values as they are typed for write-only fields.

    Omitting these keywords has the same behavior as values of false.
    """

    examples: list[Any] | None = None
    """The value of this must be an array containing the example values."""

    discriminator: Discriminator | None = None
    """Adds support for polymorphism.

    The discriminator is an object name that is used to differentiate between other schemas which may satisfy the
    payload description. See `Composition and Inheritance <https://spec.openapis.org/oas/v3.1.0#schemaComposition>`_
    for more details.
    """

    xml: XML | None = None
    """This MAY be used only on properties schemas.

    It has no effect on root schemas. Adds additional metadata to describe the XML representation of this property.
    """

    external_docs: ExternalDocumentation | None = None
    """Additional external documentation for this schema."""

    example: Any | None = None
    """A free-form property to include an example of an instance for this schema. To represent examples that cannot be
    naturally represented in JSON or YAML, a string value can be used to contain the example with escaping where
    necessary.

    Deprecated: The example property has been deprecated in favor of the JSON Schema examples keyword. Use of example is
    discouraged, and later versions of this specification may remove it.
    """

    def __hash__(self) -> int:
        return _recursive_hash(self)


@dataclass
class SchemaDataContainer(Schema):
    """Special class that allows using python data containers, e.g. dataclasses or pydantic models, to represent a
    schema
    """

    data_container: Any = None
    """A data container instance that will be used to generate the schema."""



from starlette.responses import JSONResponse

from litestar import get
from litestar.status_codes import HTTP_200_OK
from litestar.testing import create_test_client


def test_starlette_json_response() -> None:
    @get("/starlette-json-response")
    def get_json_response() -> JSONResponse:
        return JSONResponse(content={"hello": "world"})

    with create_test_client(get_json_response) as client:
        response = client.get("/starlette-json-response")
        assert response.status_code == HTTP_200_OK
        assert response.json() == {"hello": "world"}



from typing import Optional, Type

import pytest

from litestar import Controller, HttpMethod, Litestar, Response, Router, get
from litestar.handlers.http_handlers.base import HTTPRouteHandler

RouterResponse: Type[Response] = type("RouterResponse", (Response,), {})
ControllerResponse: Type[Response] = type("ControllerResponse", (Response,), {})
AppResponse: Type[Response] = type("AppResponse", (Response,), {})
HandlerResponse: Type[Response] = type("HandlerResponse", (Response,), {})


@pytest.mark.parametrize(
    "handler_response_class, controller_response_class, router_response_class, app_response_class, expected",
    (
        (HandlerResponse, ControllerResponse, RouterResponse, AppResponse, HandlerResponse),
        (None, ControllerResponse, RouterResponse, AppResponse, ControllerResponse),
        (None, None, RouterResponse, AppResponse, RouterResponse),
        (None, None, None, AppResponse, AppResponse),
        (None, None, None, None, Response),
    ),
    ids=(
        "Custom class for all layers",
        "Custom class for all above handler layer",
        "Custom class for all above controller layer",
        "Custom class for all above router layer",
        "No custom class for layers",
    ),
)
def test_response_class_resolution_of_layers(
    handler_response_class: Optional[Type[Response]],
    controller_response_class: Optional[Type[Response]],
    router_response_class: Optional[Type[Response]],
    app_response_class: Optional[Type[Response]],
    expected: Type[Response],
) -> None:
    class MyController(Controller):
        @get()
        def handler(self) -> None:
            pass

    if controller_response_class:
        MyController.response_class = ControllerResponse

    router = Router(path="/", route_handlers=[MyController])

    if router_response_class:
        router.response_class = router_response_class

    app = Litestar(route_handlers=[router])

    if app_response_class:
        app.response_class = app_response_class

    route_handler: HTTPRouteHandler = app.route_handler_method_map["/"][HttpMethod.GET]  # type: ignore[assignment]

    if handler_response_class:
        route_handler.response_class = handler_response_class

    response_class = route_handler.resolve_response_class()
    assert response_class is expected



from __future__ import annotations

from dataclasses import asdict
from typing import Callable

import pytest
from _pytest.fixtures import FixtureRequest
from fsspec.implementations.local import LocalFileSystem
from typing_extensions import TypeAlias

from litestar import Router
from litestar.file_system import BaseLocalFileSystem
from litestar.static_files import StaticFilesConfig, create_static_files_router
from litestar.types import FileSystemProtocol

MakeConfig: TypeAlias = "Callable[[StaticFilesConfig], tuple[list[StaticFilesConfig], list[Router]]]"


@pytest.fixture(params=["config", "handlers"])
def make_config(request: FixtureRequest) -> MakeConfig:
    def make(config: StaticFilesConfig) -> tuple[list[StaticFilesConfig], list[Router]]:
        if request.param == "config":
            return [config], []
        return [], [create_static_files_router(**asdict(config))]

    return make


@pytest.fixture(params=[BaseLocalFileSystem(), LocalFileSystem()])
def file_system(request: FixtureRequest) -> FileSystemProtocol:
    return request.param  # type: ignore[no-any-return]



try:
    from advanced_alchemy.exceptions import IntegrityError as ConflictError
    from advanced_alchemy.exceptions import NotFoundError, RepositoryError
except ImportError:  # pragma: no cover
    from ._exceptions import ConflictError, NotFoundError, RepositoryError  # type: ignore[assignment]

__all__ = ("ConflictError", "NotFoundError", "RepositoryError")



from __future__ import annotations

from typing import TYPE_CHECKING, Type, Union

from typing_extensions import _TypedDictMeta  # type: ignore[attr-defined]

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

__all__ = (
    "NoneType",
    "UnionType",
    "UnionTypes",
    "TypedDictClass",
)

NoneType: type[None] = type(None)

try:
    from types import UnionType  # type: ignore[attr-defined]
except ImportError:
    UnionType: TypeAlias = Union  # type: ignore[no-redef]

UnionTypes = {UnionType, Union}
TypedDictClass: TypeAlias = Type[_TypedDictMeta]



from __future__ import annotations

from json import dumps as json_dumps
from typing import TYPE_CHECKING, Callable

import pytest
from yaml import dump as dump_yaml

from litestar.cli.commands.schema import _generate_openapi_schema
from litestar.cli.main import litestar_group as cli_command

if TYPE_CHECKING:
    from pathlib import Path
    from types import ModuleType

    from click.testing import CliRunner
    from pytest import MonkeyPatch
    from pytest_mock import MockerFixture


@pytest.mark.parametrize("filename", ("", "custom.json", "custom.yaml", "custom.yml"))
def test_openapi_schema_command(
    runner: CliRunner, mocker: MockerFixture, monkeypatch: MonkeyPatch, filename: str
) -> None:
    monkeypatch.setenv("LITESTAR_APP", "test_apps.openapi_test_app.main:app")
    mock_path_write_bytes = mocker.patch("pathlib.Path.write_bytes")
    command = "schema openapi"

    from test_apps.openapi_test_app.main import app as openapi_test_app

    assert openapi_test_app.openapi_schema
    schema = openapi_test_app.openapi_schema.to_schema()

    expected_content = json_dumps(schema, indent=4).encode()
    if filename:
        command += f" --output {filename}"
        if filename.endswith(("yaml", "yml")):
            expected_content = dump_yaml(schema, default_flow_style=False, encoding="utf-8")

    result = runner.invoke(cli_command, command)
    assert result.exit_code == 0
    mock_path_write_bytes.assert_called_once_with(expected_content)


@pytest.mark.parametrize("suffix", ("json", "yaml", "yml"))
def test_schema_export_with_examples(suffix: str, create_module: Callable[[str], ModuleType], tmp_path: Path) -> None:
    module = create_module(
        """
from datetime import datetime
from litestar import Litestar, get
from litestar.openapi import OpenAPIConfig

@get()
async def something(date: datetime) -> None:
    return None

app = Litestar([something], openapi_config=OpenAPIConfig('example', '0.0.1', True))
    """
    )
    pth = tmp_path / f"openapi.{suffix}"
    _generate_openapi_schema(module.app, pth)
    assert pth.read_text()


@pytest.mark.parametrize(
    "namespace, filename", (("Custom", ""), ("", "custom_specs.ts"), ("Custom", "custom_specs.ts"))
)
def test_openapi_typescript_command(
    runner: CliRunner, mocker: MockerFixture, monkeypatch: MonkeyPatch, filename: str, namespace: str
) -> None:
    monkeypatch.setenv("LITESTAR_APP", "test_apps.openapi_test_app.main:app")
    mock_path_write_text = mocker.patch("pathlib.Path.write_text")
    command = "schema typescript"

    if namespace:
        command += f" --namespace {namespace}"
    if filename:
        command += f" --output {filename}"

    result = runner.invoke(cli_command, command)
    assert result.exit_code == 0
    assert mock_path_write_text.called


@pytest.mark.parametrize(
    "namespace, filename", (("Custom", ""), ("", "custom_specs.ts"), ("Custom", "custom_specs.ts"))
)
def test_openapi_typescript_command_without_jsbeautifier(
    runner: CliRunner, mocker: MockerFixture, monkeypatch: MonkeyPatch, filename: str, namespace: str
) -> None:
    monkeypatch.setenv("LITESTAR_APP", "test_apps.openapi_test_app.main:app")
    mocker.patch("litestar.cli.commands.schema.JSBEAUTIFIER_INSTALLED", False)
    mock_path_write_text = mocker.patch("pathlib.Path.write_text")
    command = "schema typescript"

    if namespace:
        command += f" --namespace {namespace}"
    if filename:
        command += f" --output {filename}"

    result = runner.invoke(cli_command, command)
    assert result.exit_code == 0
    assert mock_path_write_text.called



from __future__ import annotations

import dataclasses
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, cast
from unittest.mock import MagicMock

import pytest
from typing_extensions import TypeAlias

from litestar import Controller, HttpMethod, Litestar, Request, Router, delete, get
from litestar._openapi.datastructures import OpenAPIContext
from litestar._openapi.path_item import PathItemFactory, merge_path_item_operations
from litestar._openapi.utils import default_operation_id_creator
from litestar.exceptions import ImproperlyConfiguredException
from litestar.handlers.http_handlers import HTTPRouteHandler
from litestar.openapi.config import OpenAPIConfig
from litestar.openapi.spec import Operation, PathItem
from litestar.utils import find_index

if TYPE_CHECKING:
    from litestar.routes import HTTPRoute


@pytest.fixture()
def route(person_controller: type[Controller]) -> HTTPRoute:
    app = Litestar(route_handlers=[person_controller], openapi_config=None)
    index = find_index(app.routes, lambda x: x.path_format == "/{service_id}/person/{person_id}")
    return cast("HTTPRoute", app.routes[index])


@pytest.fixture()
def routes_with_router(person_controller: type[Controller]) -> tuple[HTTPRoute, HTTPRoute]:
    class PersonControllerV2(person_controller):  # type: ignore[misc, valid-type]
        pass

    router_v1 = Router(path="/v1", route_handlers=[person_controller])
    router_v2 = Router(path="/v2", route_handlers=[PersonControllerV2])
    app = Litestar(route_handlers=[router_v1, router_v2], openapi_config=None)
    index_v1 = find_index(app.routes, lambda x: x.path_format == "/v1/{service_id}/person/{person_id}")
    index_v2 = find_index(app.routes, lambda x: x.path_format == "/v2/{service_id}/person/{person_id}")
    return cast("HTTPRoute", app.routes[index_v1]), cast("HTTPRoute", app.routes[index_v2])


CreateFactoryFixture: TypeAlias = "Callable[[HTTPRoute], PathItemFactory]"


@pytest.fixture()
def create_factory() -> CreateFactoryFixture:
    def factory(route: HTTPRoute) -> PathItemFactory:
        return PathItemFactory(
            OpenAPIContext(
                openapi_config=OpenAPIConfig(title="Test", version="1.0.0", description="Test", create_examples=True),
                plugins=[],
            ),
            route,
        )

    return factory


def test_create_path_item(route: HTTPRoute, create_factory: CreateFactoryFixture) -> None:
    schema = create_factory(route).create_path_item()
    assert schema.delete
    assert schema.delete.operation_id == "ServiceIdPersonPersonIdDeletePerson"
    assert schema.delete.summary == "DeletePerson"
    assert schema.get
    assert schema.get.operation_id == "ServiceIdPersonPersonIdGetPersonById"
    assert schema.get.summary == "GetPersonById"
    assert schema.patch
    assert schema.patch.operation_id == "ServiceIdPersonPersonIdPartialUpdatePerson"
    assert schema.patch.summary == "PartialUpdatePerson"
    assert schema.put
    assert schema.put.operation_id == "ServiceIdPersonPersonIdUpdatePerson"
    assert schema.put.summary == "UpdatePerson"


def test_unique_operation_ids_for_multiple_http_methods(create_factory: CreateFactoryFixture) -> None:
    class MultipleMethodsRouteController(Controller):
        path = "/"

        @HTTPRouteHandler("/", http_method=["GET", "HEAD"])
        async def root(self, *, request: Request[str, str, Any]) -> None:
            pass

    app = Litestar(route_handlers=[MultipleMethodsRouteController], openapi_config=None)
    index = find_index(app.routes, lambda x: x.path_format == "/")
    route_with_multiple_methods = cast("HTTPRoute", app.routes[index])
    schema = create_factory(route_with_multiple_methods).create_path_item()
    assert schema.get
    assert schema.get.operation_id
    assert schema.head
    assert schema.head.operation_id
    assert schema.get.operation_id != schema.head.operation_id


def test_unique_operation_ids_for_multiple_http_methods_with_handler_level_operation_creator(
    create_factory: CreateFactoryFixture,
) -> None:
    class MultipleMethodsRouteController(Controller):
        path = "/"

        @HTTPRouteHandler("/", http_method=["GET", "HEAD"], operation_id=default_operation_id_creator)
        async def root(self, *, request: Request[str, str, Any]) -> None:
            pass

    app = Litestar(route_handlers=[MultipleMethodsRouteController], openapi_config=None)
    index = find_index(app.routes, lambda x: x.path_format == "/")
    route_with_multiple_methods = cast("HTTPRoute", app.routes[index])
    factory = create_factory(route_with_multiple_methods)
    factory.context.openapi_config.operation_id_creator = lambda x: "abc"  # type: ignore[assignment, misc]
    schema = create_factory(route_with_multiple_methods).create_path_item()
    assert schema.get
    assert schema.get.operation_id
    assert schema.head
    assert schema.head.operation_id
    assert schema.get.operation_id != schema.head.operation_id


def test_routes_with_different_paths_should_generate_unique_operation_ids(
    routes_with_router: tuple[HTTPRoute, HTTPRoute], create_factory: CreateFactoryFixture
) -> None:
    route_v1, route_v2 = routes_with_router
    schema_v1 = create_factory(route_v1).create_path_item()
    schema_v2 = create_factory(route_v2).create_path_item()
    assert schema_v1.get
    assert schema_v2.get
    assert schema_v1.get.operation_id != schema_v2.get.operation_id


def test_create_path_item_use_handler_docstring_false(route: HTTPRoute, create_factory: CreateFactoryFixture) -> None:
    factory = create_factory(route)
    assert not factory.context.openapi_config.use_handler_docstrings
    schema = factory.create_path_item()
    assert schema.get
    assert schema.get.description is None
    assert schema.patch
    assert schema.patch.description == "Description in decorator"


def test_create_path_item_use_handler_docstring_true(route: HTTPRoute, create_factory: CreateFactoryFixture) -> None:
    factory = create_factory(route)
    factory.context.openapi_config.use_handler_docstrings = True
    schema = factory.create_path_item()
    assert schema.get
    assert schema.get.description == "Description in docstring."
    assert schema.patch
    assert schema.patch.description == "Description in decorator"
    assert schema.put
    assert schema.put.description
    # make sure multiline docstring is fully included
    assert "Line 3." in schema.put.description
    # make sure internal docstring indentation used to line up with the code
    # is removed from description
    assert "    " not in schema.put.description


def test_operation_id_validation() -> None:
    @get(path="/1", operation_id="handler")
    def handler_1() -> None: ...

    @get(path="/2", operation_id="handler")
    def handler_2() -> None: ...

    app = Litestar(route_handlers=[handler_1, handler_2])

    with pytest.raises(ImproperlyConfiguredException):
        app.openapi_schema


def test_operation_override() -> None:
    @dataclass
    class CustomOperation(Operation):
        x_code_samples: list[dict[str, str]] | None = field(default=None, metadata={"alias": "x-codeSamples"})

        def __post_init__(self) -> None:
            self.tags = ["test"]
            self.description = "test"
            self.x_code_samples = [
                {"lang": "Python", "source": "import requests; requests.get('localhost/example')", "label": "Python"},
                {"lang": "cURL", "source": "curl -XGET localhost/example", "label": "curl"},
            ]

    @get(path="/1")
    def handler_1() -> None: ...

    @get(path="/2", operation_class=CustomOperation)
    def handler_2() -> None: ...

    app = Litestar(route_handlers=[handler_1, handler_2])

    assert app.openapi_schema.paths
    assert app.openapi_schema.paths["/1"]
    assert app.openapi_schema.paths["/1"].get
    assert isinstance(app.openapi_schema.paths["/1"].get, Operation)
    assert app.openapi_schema.paths["/2"]
    assert app.openapi_schema.paths["/2"].get
    assert isinstance(app.openapi_schema.paths["/2"].get, CustomOperation)
    assert app.openapi_schema.paths["/2"].get.tags == ["test"]
    assert app.openapi_schema.paths["/2"].get.description == "test"

    operation_schema = CustomOperation().to_schema()
    assert "x-codeSamples" in operation_schema


def test_handler_excluded_from_schema(create_factory: CreateFactoryFixture) -> None:
    @get("/", sync_to_thread=False)
    def handler_1() -> None: ...

    @delete("/", include_in_schema=False, sync_to_thread=False)
    def handler_2() -> None: ...

    app = Litestar(route_handlers=[handler_1, handler_2])
    index = find_index(app.routes, lambda x: x.path_format == "/")
    route_with_multiple_methods = cast("HTTPRoute", app.routes[index])
    factory = create_factory(route_with_multiple_methods)
    schema = factory.create_path_item()
    assert schema.get
    assert schema.delete is None


@pytest.mark.parametrize("method", HttpMethod)
def test_merge_path_item_operations_operation_set_on_both_raises(method: HttpMethod) -> None:
    with pytest.raises(ValueError, match="Cannot merge operation"):
        merge_path_item_operations(
            PathItem(**{method.value.lower(): MagicMock()}),
            PathItem(**{method.value.lower(): MagicMock()}),
            for_path="/",
        )


@pytest.mark.parametrize(
    "attr",
    [
        f.name
        for f in dataclasses.fields(PathItem)
        if f.name.upper()
        not in [
            *HttpMethod,
            "TRACE",  # remove once https://github.com/litestar-org/litestar/pull/3294 is merged
        ]
    ],
)
def test_merge_path_item_operation_differing_values_raises(attr: str) -> None:
    with pytest.raises(ImproperlyConfiguredException, match="Conflicting OpenAPI path configuration for '/'"):
        merge_path_item_operations(PathItem(), PathItem(**{attr: MagicMock()}), for_path="/")



from __future__ import annotations

from functools import wraps
from inspect import Parameter, Signature
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Dict

from msgspec.json import Encoder as JsonEncoder

from litestar.di import Provide
from litestar.serialization import decode_json
from litestar.types.builtin_types import NoneType
from litestar.utils import ensure_async_callable
from litestar.utils.helpers import unwrap_partial

if TYPE_CHECKING:
    from litestar import WebSocket
    from litestar.handlers.websocket_handlers.listener import WebsocketListenerRouteHandler
    from litestar.types import AnyCallable
    from litestar.utils.signature import ParsedSignature


def create_handle_receive(listener: WebsocketListenerRouteHandler) -> Callable[[WebSocket], Coroutine[Any, None, None]]:
    if data_dto := listener.resolve_data_dto():

        async def handle_receive(socket: WebSocket) -> Any:
            received_data = await socket.receive_data(mode=listener._receive_mode)
            return data_dto(socket).decode_bytes(
                received_data.encode("utf-8") if isinstance(received_data, str) else received_data
            )

    elif listener.parsed_data_field and listener.parsed_data_field.annotation is str:

        async def handle_receive(socket: WebSocket) -> Any:
            received_data = await socket.receive_data(mode=listener._receive_mode)
            return received_data.decode("utf-8") if isinstance(received_data, bytes) else received_data

    elif listener.parsed_data_field and listener.parsed_data_field.annotation is bytes:

        async def handle_receive(socket: WebSocket) -> Any:
            received_data = await socket.receive_data(mode=listener._receive_mode)
            return received_data.encode("utf-8") if isinstance(received_data, str) else received_data

    else:

        async def handle_receive(socket: WebSocket) -> Any:
            received_data = await socket.receive_data(mode=listener._receive_mode)
            return decode_json(value=received_data, type_decoders=socket.route_handler.resolve_type_decoders())

    return handle_receive


def create_handle_send(
    listener: WebsocketListenerRouteHandler,
) -> Callable[[WebSocket, Any], Coroutine[None, None, None]]:
    json_encoder = JsonEncoder(enc_hook=listener.default_serializer)

    if return_dto := listener.resolve_return_dto():

        async def handle_send(socket: WebSocket, data: Any) -> None:
            encoded_data = return_dto(socket).data_to_encodable_type(data)
            data = json_encoder.encode(encoded_data)
            await socket.send_data(data=data, mode=listener._send_mode)

    elif listener.parsed_return_field.is_subclass_of((str, bytes)) or (
        listener.parsed_return_field.is_optional and listener.parsed_return_field.has_inner_subclass_of((str, bytes))
    ):

        async def handle_send(socket: WebSocket, data: Any) -> None:
            await socket.send_data(data=data, mode=listener._send_mode)

    else:

        async def handle_send(socket: WebSocket, data: Any) -> None:
            data = json_encoder.encode(data)
            await socket.send_data(data=data, mode=listener._send_mode)

    return handle_send


class ListenerHandler:
    __slots__ = ("_can_send_data", "_fn", "_listener", "_pass_socket")

    def __init__(
        self,
        listener: WebsocketListenerRouteHandler,
        fn: AnyCallable,
        parsed_signature: ParsedSignature,
        namespace: dict[str, Any],
    ) -> None:
        self._can_send_data = not parsed_signature.return_type.is_subclass_of(NoneType)
        self._fn = ensure_async_callable(fn)
        self._listener = listener
        self._pass_socket = "socket" in parsed_signature.parameters

    async def __call__(
        self,
        *args: Any,
        socket: WebSocket,
        connection_lifespan_dependencies: Dict[str, Any],  # noqa: UP006
        **kwargs: Any,
    ) -> None:
        lifespan_mananger = self._listener._connection_lifespan or self._listener.default_connection_lifespan
        handle_send = self._listener.resolve_send_handler() if self._can_send_data else None
        handle_receive = self._listener.resolve_receive_handler()

        if self._pass_socket:
            kwargs["socket"] = socket

        async with lifespan_mananger(**connection_lifespan_dependencies):
            while True:
                received_data = await handle_receive(socket)
                data = await self._fn(*args, data=received_data, **kwargs)
                if handle_send:
                    await handle_send(socket, data)


def create_handler_signature(callback_signature: Signature) -> Signature:
    """Creates a :class:`Signature` for the handler function for signature modelling.

    This is required for two reasons:

        1. the :class:`.handlers.WebsocketHandler` signature model cannot contain the ``data`` parameter, which is
            required for :class:`.handlers.websocket_listener` handlers.
        2. the :class;`.handlers.WebsocketHandler` signature model must include the ``socket`` parameter, which is
            optional for :class:`.handlers.websocket_listener` handlers.

    Args:
        callback_signature: The :class:`Signature` of the listener callback.

    Returns:
        The :class:`Signature` for the listener callback as required for signature modelling.
    """
    new_params = [p for p in callback_signature.parameters.values() if p.name != "data"]
    if "socket" not in callback_signature.parameters:
        new_params.append(Parameter(name="socket", kind=Parameter.KEYWORD_ONLY, annotation="WebSocket"))

    new_params.append(
        Parameter(name="connection_lifespan_dependencies", kind=Parameter.KEYWORD_ONLY, annotation="Dict[str, Any]")
    )

    return callback_signature.replace(parameters=new_params)


def create_stub_dependency(src: AnyCallable) -> Provide:
    """Create a stub dependency, accepting any kwargs defined in ``src``, and
    wrap it in ``Provide``
    """
    src = unwrap_partial(src)

    @wraps(src)
    async def stub(**kwargs: Any) -> Dict[str, Any]:  # noqa: UP006
        return kwargs

    return Provide(stub)



from ._async import SQLAlchemyAsyncRepository
from ._sync import SQLAlchemySyncRepository
from ._util import wrap_sqlalchemy_exception
from .types import ModelT

__all__ = (
    "SQLAlchemyAsyncRepository",
    "SQLAlchemySyncRepository",
    "ModelT",
    "wrap_sqlalchemy_exception",
)



from __future__ import annotations

from collections import defaultdict
from functools import lru_cache
from http.cookies import _unquote as unquote_cookie
from urllib.parse import unquote

try:
    from fast_query_parsers import parse_query_string as parse_qsl
except ImportError:
    from urllib.parse import parse_qsl as _parse_qsl

    def parse_qsl(qs: bytes, separator: str) -> list[tuple[str, str]]:
        return _parse_qsl(qs.decode("latin-1"), keep_blank_values=True, separator=separator)


__all__ = ("parse_cookie_string", "parse_query_string", "parse_url_encoded_form_data")


@lru_cache(1024)
def parse_url_encoded_form_data(encoded_data: bytes) -> dict[str, str | list[str]]:
    """Parse an url encoded form data dict.

    Args:
        encoded_data: The encoded byte string.

    Returns:
        A parsed dict.
    """
    decoded_dict: defaultdict[str, list[str]] = defaultdict(list)
    for k, v in parse_qsl(encoded_data, separator="&"):
        decoded_dict[k].append(v)
    return {k: v if len(v) > 1 else v[0] for k, v in decoded_dict.items()}


@lru_cache(1024)
def parse_query_string(query_string: bytes) -> tuple[tuple[str, str], ...]:
    """Parse a query string into a tuple of key value pairs.

    Args:
        query_string: A query string.

    Returns:
        A tuple of key value pairs.
    """
    return tuple(parse_qsl(query_string, separator="&"))


@lru_cache(1024)
def parse_cookie_string(cookie_string: str) -> dict[str, str]:
    """Parse a cookie string into a dictionary of values.

    Args:
        cookie_string: A cookie string.

    Returns:
        A string keyed dictionary of values
    """
    cookies = [cookie.split("=", 1) if "=" in cookie else ("", cookie) for cookie in cookie_string.split(";")]
    output: dict[str, str] = {
        k: unquote(unquote_cookie(v))
        for k, v in filter(
            lambda x: x[0] or x[1],
            ((k.strip(), v.strip()) for k, v in cookies),
        )
    }
    return output



from __future__ import annotations

import atexit
import inspect
import logging
import random
import sys
from contextlib import AbstractContextManager, contextmanager
from typing import Any, AsyncContextManager, Awaitable, ContextManager, Generator, TypeVar, cast, overload

import picologging
from _pytest.logging import LogCaptureHandler, _LiveLoggingNullHandler

from litestar._openapi.schema_generation import SchemaCreator
from litestar._openapi.schema_generation.plugins import openapi_schema_plugins
from litestar.openapi.spec import Schema
from litestar.plugins import OpenAPISchemaPluginProtocol
from litestar.typing import FieldDefinition

T = TypeVar("T")


RANDOM = random.Random(b"bA\xcd\x00\xa9$\xa7\x17\x1c\x10")


# TODO: Remove when dropping 3.9
if sys.version_info < (3, 9):

    def randbytes(n: int) -> bytes:
        return RANDOM.getrandbits(8 * n).to_bytes(n, "little")

else:
    randbytes = RANDOM.randbytes


if sys.version_info >= (3, 12):
    getHandlerByName = logging.getHandlerByName
else:
    from logging import _handlers  # type: ignore[attr-defined]

    def getHandlerByName(name: str) -> Any:
        return _handlers.get(name)


@overload
async def maybe_async(obj: Awaitable[T]) -> T: ...


@overload
async def maybe_async(obj: T) -> T: ...


async def maybe_async(obj: Awaitable[T] | T) -> T:
    return cast(T, await obj) if inspect.isawaitable(obj) else cast(T, obj)


class _AsyncContextManagerWrapper(AsyncContextManager):
    def __init__(self, cm: AbstractContextManager):
        self.cm = cm

    async def __aenter__(self) -> Any:
        return self.cm.__enter__()

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> Any:
        return self.cm.__exit__(exc_type, exc_val, exc_tb)


def maybe_async_cm(obj: ContextManager[T] | AsyncContextManager[T]) -> AsyncContextManager[T]:
    if isinstance(obj, AbstractContextManager):
        return cast(AsyncContextManager[T], _AsyncContextManagerWrapper(obj))
    return obj


def get_schema_for_field_definition(
    field_definition: FieldDefinition, *, plugins: list[OpenAPISchemaPluginProtocol] | None = None
) -> Schema:
    plugins = [*openapi_schema_plugins, *(plugins or [])]
    creator = SchemaCreator(plugins=plugins)
    result = creator.for_field_definition(field_definition)
    if isinstance(result, Schema):
        return result
    return creator.schema_registry.from_reference(result).schema


@contextmanager
def cleanup_logging_impl() -> Generator:
    # Reset root logger (`logging` module)
    std_root_logger: logging.Logger = logging.getLogger()
    for std_handler in std_root_logger.handlers:
        # Don't interfere with PyTest handler config
        if not isinstance(std_handler, (_LiveLoggingNullHandler, LogCaptureHandler)):
            std_root_logger.removeHandler(std_handler)

    # Reset root logger (`picologging` module)
    pico_root_logger: picologging.Logger = picologging.getLogger()
    for pico_handler in pico_root_logger.handlers:
        pico_root_logger.removeHandler(pico_handler)

    yield

    # Stop queue_listener listener (mandatory for the 'logging' module with Python 3.12,
    # else the test suite would hang on at the end of the tests and some tests would fail)
    queue_listener_handler = getHandlerByName("queue_listener")
    if queue_listener_handler and hasattr(queue_listener_handler, "listener"):
        atexit.unregister(queue_listener_handler.listener.stop)
        queue_listener_handler.listener.stop()
        queue_listener_handler.close()
        del queue_listener_handler
